__all__ = ["HydraulicLoopDialog"]

import dataclasses as _dc
import itertools as _it
import typing as _tp

import PyQt5.QtCore as _qtc
import PyQt5.QtGui as _qtg
from PyQt5 import QtWidgets as _qtw

try:
    from . import _UI_dialog_generated as _uigen
    import trnsysGUI.resources.QRC_resources_generated as _
except ImportError as importError:
    raise AssertionError(  # pylint: disable=duplicate-code  # 1
        "Could not find the generated Python code for a .ui or .qrc file. Please run the "
        "`dev-tools\\generateGuiClassesFromQtCreatorStudioUiFiles.py' script from your "
        "`pytrnsys_gui` directory."
    ) from importError

import trnsysGUI.hydraulicLoops.model as _model

from . import model as _gmodel


class HydraulicLoopDialog(_qtw.QDialog, _uigen.Ui_hydraulicLoopDialog):
    def __init__(self, hydraulicLoop: _gmodel.HydraulicLoop, occupiedNames: _tp.Sequence[str], fluids: _model.Fluids):
        super().__init__()
        self.setupUi(self)

        self._hydraulicLoop = hydraulicLoop
        self._occupiedNames = occupiedNames
        self._fluids = fluids

        self._configureAndInitializeLoopGroupBox()

        self._configureBulkOperations()

        self._reloadConnections()

    def _configureAndInitializeLoopGroupBox(self) -> None:
        self._configureLoopNameLineEdit()
        self.loopName.setText(self._hydraulicLoop.name)

        self._configureFluidComboBox()
        fluidIndex = self.fluidComboBox.findData(self._hydraulicLoop.fluid)
        assert fluidIndex >= 0
        self.fluidComboBox.setCurrentIndex(fluidIndex)

        self._setUseLoopWideDefaults(self._hydraulicLoop.useLoopWideDefaults)
        self._configureUseLoopWideDefaultsCheckbox()

    def _configureUseLoopWideDefaultsCheckbox(self) -> None:
        def onToggled(isSet: bool) -> None:
            if isSet and not self._doesUserSayContinueWithEnablingLoopWideDefaults():
                self.useLoopWideDefaultsCheckBox.setChecked(False)
                return

            self._setUseLoopWideDefaults(isSet)

            self.adjustSize()

        self.useLoopWideDefaultsCheckBox.toggled.connect(onToggled)

    def _configureLoopNameLineEdit(self) -> None:
        def onNameChanged(newName: str) -> None:
            self._hydraulicLoop.name = newName

            isNameEmpty = not newName
            isNameOccupied = self._isNameOccupied(newName)

            if isNameEmpty:
                tooltipText = "You must specify a name"
            elif isNameOccupied:
                tooltipText = "This name is already in use"
            else:
                tooltipText = ""
            self.invalidNameWarningWidget.setToolTip(tooltipText)

            haveError = isNameEmpty or isNameOccupied
            self.invalidNameWarningWidget.setVisible(haveError)

            okButton = self.okCancelButtonBox.button(self.okCancelButtonBox.Ok)
            okButton.setEnabled(not haveError)

        self.loopName.textChanged.connect(onNameChanged)

    def _isNameOccupied(self, newName: str) -> bool:
        return newName in self._occupiedNames

    def _configureFluidComboBox(self) -> None:
        for fluid in self._fluids.fluids:
            self.fluidComboBox.addItem(fluid.name, userData=fluid)

        def onCurrentIndexChanged(_) -> None:
            self._hydraulicLoop.fluid = self.fluidComboBox.currentData()

        self.fluidComboBox.currentIndexChanged.connect(onCurrentIndexChanged)

    def _configureBulkOperations(self) -> None:
        def setTotalLoopLengthInfoToolTipVisible(_) -> None:
            isVisible = self._isBulkTotalPipeLengthTypeSelected()
            self.bulkLengthInfoIconWidget.setVisible(isVisible)

        self.bulkLengthTypeComboBox.currentTextChanged.connect(setTotalLoopLengthInfoToolTipVisible)

        def onBulkApplyButtonPressed() -> None:
            individualPipeLengthInM = self._bulkIndividualPipeLengthInM

            diameterText = self.bulkDiameterLineEdit.text()
            diameterInCm = float(diameterText) if diameterText else None

            uValueText = self.bulkUValueLineEdit.text()
            uValueInWPerM2K = float(uValueText) if uValueText else None

            for connection in self._hydraulicLoop.connections:
                if individualPipeLengthInM is not None:
                    connection.lengthInM = individualPipeLengthInM

                if diameterInCm is not None:
                    connection.diameterInCm = diameterInCm

                if uValueInWPerM2K is not None:
                    connection.uValueInWPerM2K = uValueInWPerM2K

            self._reloadConnections()

        def clearLineEdits() -> None:
            self.bulkLengthLineEdit.clear()
            self.bulkDiameterLineEdit.clear()
            self.bulkUValueLineEdit.clear()

        self.bulkApplyButton.pressed.connect(onBulkApplyButtonPressed)
        self.bulkApplyButton.pressed.connect(clearLineEdits)

        self.bulkCancelButton.pressed.connect(clearLineEdits)

        self._configureForPositiveFloat(self.bulkLengthLineEdit)
        self._configureForPositiveFloat(self.bulkDiameterLineEdit)
        self._configureForPositiveFloat(self.bulkUValueLineEdit)

    def _isBulkTotalPipeLengthTypeSelected(self) -> bool:
        text = self.bulkLengthTypeComboBox.currentText()

        if text == "Individual pipe length":
            return False

        if text == "Total loop length":
            return True

        raise AssertionError(f"Unknown bulk length type text: {text}")

    @staticmethod
    def _configureForPositiveFloat(lineEdit: _qtw.QLineEdit) -> None:
        def checkIfTextIsPositiveFloatOrReset() -> None:
            text = lineEdit.text()
            try:
                _parsePositiveFloat(text)
            except ValueError:
                lineEdit.setText("")

        lineEdit.editingFinished.connect(checkIfTextIsPositiveFloatOrReset)

    def _reloadConnections(self) -> None:
        connectionsModel = _ConnectionsUiModel(self._hydraulicLoop.connections)

        connectionsModel.dataChanged.connect(lambda *_: self._updatePipesStatisticsLabel())

        self.connectionsTableView.setModel(connectionsModel)

        self.connectionsTableView.resizeRowsToContents()
        self.connectionsTableView.resizeColumnsToContents()

        self._updatePipesStatisticsLabel()

    def _updatePipesStatisticsLabel(self) -> None:
        totalLoopLengthInM = sum(c.lengthInM for c in self._hydraulicLoop.connections)
        statisticsLabelText = (
            f"Number of pipes: {len(self._hydraulicLoop.connections)}, total loop length: {totalLoopLengthInM:g} m"
        )
        self.pipesStatisticsLabel.setText(statisticsLabelText)

    @property
    def _bulkIndividualPipeLengthInM(self) -> _tp.Optional[float]:
        lengthText = self.bulkLengthLineEdit.text()
        if not lengthText:
            return None

        haveTotalPipeLength = self._isBulkTotalPipeLengthTypeSelected()

        if not haveTotalPipeLength:
            individualPipeLengthInM = float(lengthText)
            return individualPipeLengthInM

        loopLengthInM = float(lengthText)
        nPipes = len(self._hydraulicLoop.connections)
        individualPipeLengthM = loopLengthInM / nPipes
        return individualPipeLengthM

    def _doesUserSayContinueWithEnablingLoopWideDefaults(self):
        messageBox = _qtw.QMessageBox(self)
        messageBox.setWindowTitle("Do you want to proceed?")
        messageBox.setText(
            "The values of the length, diameter and U-Value for every pipe in the loop will be "
            "lost upon OK'ing the main dialog when enabling loop wide defaults. Do you want to proceed?"
        )
        messageBox.setStandardButtons(messageBox.Yes | messageBox.No)
        pressedButton = messageBox.exec()
        shallContinue = (pressedButton == messageBox.Yes)
        return shallContinue

    def _setUseLoopWideDefaults(self, isSet: bool) -> None:
        self._hydraulicLoop.useLoopWideDefaults = isSet
        self.useLoopWideDefaultsCheckBox.setChecked(isSet)
        isEnabled = not isSet
        self.bulkOperationsGroupBox.setVisible(isEnabled)
        self.pipesGroupBox.setVisible(isEnabled)
        self.adjustSize()

    @classmethod
    def showDialog(
        cls, hydraulicLoop: _gmodel.HydraulicLoop, occupiedNames: _tp.Sequence[str], fluids: _model.Fluids
    ) -> _tp.Literal["oked", "cancelled"]:
        dialog = HydraulicLoopDialog(hydraulicLoop, occupiedNames, fluids)

        dialogCode = dialog.exec()

        if dialogCode == _qtw.QDialog.Accepted:
            return "oked"

        if dialogCode == _qtw.QDialog.Rejected:
            return "cancelled"

        raise AssertionError(f"Unknown dialog code {dialogCode}.")


_PropertyValue = _tp.Union[str, float]


@_dc.dataclass
class _Property:
    header: str
    getter: _tp.Callable[[_gmodel.Connection], _PropertyValue]
    setter: _tp.Optional[_tp.Callable[[_gmodel.Connection, _PropertyValue], None]] = None
    shallHighlightOutliers: bool = True


def _parsePositiveFloat(text: str) -> float:
    value = float(text)

    if value < 0:
        raise ValueError("Value must be positive.")

    return value


def _setConnectionDiameter(connection: _gmodel.Connection, diameterInCmText: _tp.Any) -> None:
    connection.diameterInCm = _parsePositiveFloat(diameterInCmText)


def _setConnectionUValue(connection: _gmodel.Connection, uValueInWPerM2KText: _tp.Any) -> None:
    connection.uValueInWPerM2K = _parsePositiveFloat(uValueInWPerM2KText)


def _setConnectionLength(connection: _gmodel.Connection, lengthInMText: _tp.Any) -> None:
    connection.lengthInM = _parsePositiveFloat(lengthInMText)


class _ConnectionsUiModel(_qtc.QAbstractItemModel):
    c: _gmodel.Connection
    _PROPERTIES = [
        _Property("Name", lambda c: c.name, shallHighlightOutliers=False),
        _Property("Diameter [cm]", lambda c: c.diameterInCm, _setConnectionDiameter),
        _Property("U value [W/(m^2 K)]", lambda c: c.uValueInWPerM2K, _setConnectionUValue),
        _Property("Length [m]", lambda c: c.lengthInM, _setConnectionLength),
    ]

    def __init__(self, connections: _tp.Sequence[_gmodel.Connection]):
        super().__init__()

        self.connections = connections

    def rowCount(self, parent: _qtc.QModelIndex = _qtc.QModelIndex()) -> int:
        if not self._isTopLevel(parent):
            return 0

        return len(self.connections)

    def columnCount(self, parent: _qtc.QModelIndex = _qtc.QModelIndex()) -> int:
        if not self._isTopLevel(parent):
            return 0

        return len(self._headers)

    def headerData(self, section: int, orientation: _qtc.Qt.Orientation, role: int = _qtc.Qt.DisplayRole) -> _tp.Any:
        if role != _qtc.Qt.DisplayRole or orientation != _qtc.Qt.Horizontal:
            return None

        return self._headers[section]

    @property
    def _headers(self) -> _tp.Sequence[str]:
        return [p.header for p in self._PROPERTIES]

    def data(self, modelIndex: _qtc.QModelIndex, role: int = _qtc.Qt.DisplayRole) -> _tp.Any:
        connection, prop = self._getConnectionAndProperty(modelIndex)

        if role == _qtc.Qt.DisplayRole:
            data = prop.getter(connection)  # type: ignore[misc,call-arg]
            return data

        if prop.shallHighlightOutliers and role in [_qtc.Qt.FontRole, _qtc.Qt.ForegroundRole]:
            return self._getHighlightData(connection, prop, role)

        return None

    def _getHighlightData(self, connection: _gmodel.Connection, prop: _Property, role: int) -> _tp.Any:
        mostUsedValue = self._getMostUsedValue(prop)
        value = prop.getter(connection)  # type: ignore[misc,call-arg]
        shallHighlight = value != mostUsedValue

        if role == _qtc.Qt.FontRole:
            if not shallHighlight:
                return _qtg.QFont()

            highlightFont = _qtg.QFont()
            highlightFont.setBold(True)
            return highlightFont

        if role == _qtc.Qt.ForegroundRole:
            if not shallHighlight:
                return _qtg.QColorConstants.Black

            return _qtg.QColorConstants.Red

        raise ValueError(f"Role must be {_qtc.Qt.FontRole} or {_qtc.Qt.ForegroundRole}.", role)

    def _getMostUsedValue(self, prop: _Property) -> _PropertyValue:
        sortedValues = sorted(prop.getter(c) for c in self.connections)  # type: ignore[misc, call-arg]
        groupedValues = _it.groupby(sortedValues)
        valuesWithCount = [{"value": v, "count": len(list(vs))} for v, vs in groupedValues]
        mostUsedValueWithCount = max(valuesWithCount, key=lambda vwc: vwc["count"])
        return mostUsedValueWithCount["value"]

    def setData(self, index: _qtc.QModelIndex, value: _tp.Any, role: int = _qtc.Qt.EditRole) -> bool:
        if role != _qtc.Qt.EditRole:
            return False

        connection, prop = self._getConnectionAndProperty(index)

        try:
            prop.setter(connection, value)  # type: ignore[misc]
        except ValueError:
            return False

        self.dataChanged.emit(index, index, [role])

        return True

    def _getConnectionAndProperty(self, modelIndex: _qtc.QModelIndex) -> _tp.Tuple[_gmodel.Connection, _Property]:
        rowIndex = modelIndex.row()
        connection = self.connections[rowIndex]

        propertyIndex = modelIndex.column()
        prop = self._PROPERTIES[propertyIndex]

        return connection, prop

    def flags(self, index: _qtc.QModelIndex) -> _qtc.Qt.ItemFlags:
        parentFlags = super().flags(index)

        if not self._isEditable(index.column()):
            return parentFlags

        return parentFlags | _qtc.Qt.ItemIsEditable  # type: ignore[operator,return-value]

    def _isEditable(self, columnIndex: int) -> bool:
        prop = self._PROPERTIES[columnIndex]
        return bool(prop.setter)

    def index(self, row: int, column: int, parent: _tp.Optional[_qtc.QModelIndex] = None) -> _qtc.QModelIndex:
        assert parent

        if not self._isTopLevel(parent):
            return _qtc.QModelIndex()

        if not 0 <= row < self.rowCount():
            return _qtc.QModelIndex()

        if not 0 <= column < self.columnCount():
            return _qtc.QModelIndex()

        return self.createIndex(row, column)

    def parent(self, _: _qtc.QModelIndex) -> _qtc.QModelIndex:  # type: ignore[override] # pylint: disable=no-self-use
        return _qtc.QModelIndex()

    @staticmethod
    def _isTopLevel(parent: _qtc.QModelIndex) -> bool:
        return not parent.isValid()



import PyQt5.QtWidgets as _qtw
import PyQt5.QtCore as _qtc


class EnterLeaveNotifyingGroupBox(_qtw.QGroupBox):
    enter = _qtc.pyqtSignal()
    leave = _qtc.pyqtSignal()

    def enterEvent(self, _: _qtc.QEvent) -> None:
        self.enter.emit()

    def leaveEvent(self, _: _qtc.QEvent) -> None:
        self.leave.emit()



# pylint: skip-file
# type: ignore

import os
import shutil
import typing as _tp

from PyQt5.QtWidgets import QTreeView

import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.MyQFileSystemModel import MyQFileSystemModel
from trnsysGUI.MyQTreeView import MyQTreeView
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin


class AirSourceHP(BlockItem, MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super(AirSourceHP, self).__init__(trnsysType, parent, **kwargs)

        self.inputs.append(_cspi.createSinglePipePortItem("i", 2, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))
        self.path = ""
        self.loadedFiles = []

        self.changeSize()
        self.addTree()

    def getInternalPiping(self) -> InternalPiping:
        inputPort = _mfn.PortItem("input", _mfn.PortItemType.INPUT)
        outputPort = _mfn.PortItem("output", _mfn.PortItemType.OUTPUT)
        pipe = _mfn.Pipe(self.displayName, self.trnsysId, inputPort, outputPort)

        return InternalPiping([pipe], {inputPort: self.inputs[0], outputPort: self.outputs[0]})

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.AIR_SOURCE_HP_SVG

    def changeSize(self):
        w = self.w
        h = self.h

        """ Resize block function """
        delta = 20
        deltaH = self.h / 10

        # Limit the block size:
        if h < 20:
            h = 20
        if w < 40:
            w = 40
        # center label:
        rect = self.label.boundingRect()
        lw, lh = rect.width(), rect.height()
        lx = (w - lw) / 2
        self.label.setPos(lx, h)

        self.origInputsPos = [[w, h - delta]]
        self.origOutputsPos = [[w, delta]]

        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4
        self.outputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4

        return w, h

    def addTree(self):
        """
        When a blockitem is added to the main window.
        A file explorer for that item is added to the right of the main window by calling this method
        """
        self.logger.debug(self.parent.parent())
        pathName = self.displayName
        if self.parent.parent().projectPath == "":
            self.path = self.parent.parent().projectFolder
        else:
            self.path = self.parent.parent().projectPath
        self.path = os.path.join(self.path, "ddck")
        self.path = os.path.join(self.path, pathName)
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.model = MyQFileSystemModel()
        self.model.setRootPath(self.path)
        self.model.setName(self.displayName)
        self.tree = MyQTreeView(self.model, self)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(self.path))
        self.tree.setObjectName("%sTree" % self.displayName)
        self.tree.setMinimumHeight(200)
        self.tree.setSortingEnabled(True)
        self.parent.parent().splitter.addWidget(self.tree)

    def hasDdckPlaceHolders(self):
        return True

    def deleteBlock(self):
        """
        Overridden method to also delete folder
        """
        self.logger.debug("Block " + str(self) + " is deleting itself (" + self.displayName + ")")
        self.deleteConns()
        self.parent.parent().trnsysObj.remove(self)
        self.logger.debug("deleting block " + str(self) + self.displayName)
        self.parent.scene().removeItem(self)
        widgetToRemove = self.parent.parent().findChild(QTreeView, self.displayName + "Tree")
        shutil.rmtree(self.path)
        self.deleteLoadedFile()
        try:
            widgetToRemove.hide()
        except AttributeError:
            self.logger.debug("Widget doesnt exist!")
        else:
            self.logger.debug("Deleted widget")
        del self

    def setDisplayName(self, newName):
        """
        Overridden method to also change folder name
        """
        self.displayName = newName
        self.label.setPlainText(newName)
        self.model.setName(self.displayName)
        self.tree.setObjectName("%sTree" % self.displayName)
        self.logger.debug(os.path.dirname(self.path))
        destPath = os.path.join(os.path.split(self.path)[0], self.displayName)
        if os.path.exists(self.path):
            os.rename(self.path, destPath)
            self.path = destPath
            self.logger.debug(self.path)



# pylint: skip-file
# type: ignore

import glob
import os
import shutil
import typing as _tp

from PyQt5.QtWidgets import QTreeView

import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.MyQFileSystemModel import MyQFileSystemModel
from trnsysGUI.MyQTreeView import MyQTreeView
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin


class HeatPump(BlockItem, MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))
        self.inputs.append(_cspi.createSinglePipePortItem("i", 2, self))

        self.outputs.append(_cspi.createSinglePipePortItem("o", 0, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))
        self.loadedFiles = []

        # For restoring correct order of trnsysObj list
        self.childIds = []
        self.childIds.append(self.trnsysId)
        self.childIds.append(self.parent.parent().idGen.getTrnsysID())

        self.changeSize()
        self.addTree()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.HP_SVG

    def changeSize(self):
        w = self.w
        h = self.h

        """ Resize block function """
        delta = 20

        # Limit the block size:
        if h < 20:
            h = 20
        if w < 40:
            w = 40
        # center label:
        rect = self.label.boundingRect()
        lw, lh = rect.width(), rect.height()
        lx = (w - lw) / 2
        self.label.setPos(lx, h)

        # upper left is reference
        self.origInputsPos = [[0, delta], [w, h - delta]]  # inlet of [evap, cond]
        self.origOutputsPos = [[0, h - delta], [w, delta]]  # outlet of [evap, cond]

        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])  # evaporator
        self.inputs[1].setPos(self.origInputsPos[1][0], self.origInputsPos[1][1])  # condenser

        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])
        self.outputs[1].setPos(self.origOutputsPos[1][0], self.origOutputsPos[1][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.inputs[1].side = (self.rotationN + 2 - 2 * self.flippedH) % 4

        self.outputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.outputs[1].side = (self.rotationN + 2 - 2 * self.flippedH) % 4
        return w, h

    def encode(self):
        if self.isVisible():
            self.logger.debug("Encoding a HeatPump")

            portListInputs = []
            portListOutputs = []

            for p in self.inputs:
                portListInputs.append(p.id)
            for p in self.outputs:
                portListOutputs.append(p.id)

            dct = {}
            dct[".__BlockDict__"] = True
            dct["BlockName"] = self.name
            dct["BlockDisplayName"] = self.displayName
            dct["PortsIDIn"] = portListInputs
            dct["PortsIDOut"] = portListOutputs
            dct["HeatPumpPosition"] = (float(self.pos().x()), float(self.pos().y()))
            dct["ID"] = self.id
            dct["trnsysID"] = self.trnsysId
            dct["childIds"] = self.childIds
            dct["FlippedH"] = self.flippedH
            dct["FlippedV"] = self.flippedH
            dct["RotationN"] = self.rotationN

            dictName = "Block-"

            return dictName, dct

    def decode(self, i, resBlockList):
        self.logger.debug("Loading a HeatPump block")

        self.flippedH = i["FlippedH"]
        self.flippedV = i["FlippedV"]
        self.childIds = i["childIds"]
        self.displayName = i["BlockDisplayName"]
        self.changeSize()

        for x in range(len(self.inputs)):
            self.inputs[x].id = i["PortsIDIn"][x]
            self.logger.debug("Input at heatExchanger")

        for x in range(len(self.outputs)):
            self.outputs[x].id = i["PortsIDOut"][x]
            self.logger.debug("Output at heatExchanger")

        self.setPos(float(i["HeatPumpPosition"][0]), float(i["HeatPumpPosition"][1]))
        self.trnsysId = i["trnsysID"]
        self.id = i["ID"]

        resBlockList.append(self)

    def decodePaste(self, i, offset_x, offset_y, resConnList, resBlockList, **kwargs):
        self.logger.debug("Loading a HeatPump in Decoder")

        self.changeSize()

        for x in range(len(self.inputs)):
            self.inputs[x].id = i["PortsIDIn"][x]
            self.logger.debug("Input at heatExchanger")

        for x in range(len(self.outputs)):
            self.outputs[x].id = i["PortsIDOut"][x]
            self.logger.debug("Output at heatExchanger")

        self.setPos(float(i["HeatPumpPosition"][0]) + offset_x, float(i["HeatPumpPosition"][1] + offset_y))
        resBlockList.append(self)

    def exportBlackBox(self):
        equation = []
        files = glob.glob(os.path.join(self.path, "**/*.ddck"), recursive=True)
        if not (files):
            status = "noDdckFile"
            for i in range(1, 3):
                equation.append("T" + self.displayName + "X" + str(i) + "=1")
        else:
            status = "noDdckEntry"
        lines = []
        for file in files:
            infile = open(file, "r")
            lines += infile.readlines()
        for i in range(len(lines)):
            if "output" in lines[i].lower() and "to" in lines[i].lower() and "hydraulic" in lines[i].lower():
                counter = 1
                for j in range(i, len(lines) - i):
                    if lines[j][0] == "T":
                        outputT = lines[j].split("=")[0].replace(" ", "")
                        equation.append("T" + self.displayName + "X" + str(counter) + "=1 ! suggestion: " + outputT)
                        counter += 1
                    if counter == 3:
                        status = "success"
                        break
                break

        return status, equation

    def getInternalPiping(self) -> InternalPiping:
        condenserInput = _mfn.PortItem("condenserInput", _mfn.PortItemType.INPUT)
        condenserOutput = _mfn.PortItem("condenserOutput", _mfn.PortItemType.OUTPUT)
        condenserPipe = _mfn.Pipe(f"{self.displayName}Cond", self.childIds[0], condenserInput, condenserOutput)

        evaporatorInput = _mfn.PortItem("evaporatorInput", _mfn.PortItemType.INPUT)
        evaporatorOutput = _mfn.PortItem("evaporatorOutput", _mfn.PortItemType.OUTPUT)
        evaporatorPipe = _mfn.Pipe(f"{self.displayName}Evap", self.childIds[1], evaporatorInput, evaporatorOutput)

        modelPortItemsToGraphicalPortItem = {
            condenserInput: self.inputs[1],
            condenserOutput: self.outputs[1],
            evaporatorInput: self.inputs[0],
            evaporatorOutput: self.outputs[0],
        }
        nodes = [condenserPipe, evaporatorPipe]

        return InternalPiping(nodes, modelPortItemsToGraphicalPortItem)

    def getSubBlockOffset(self, c):
        for i in range(2):
            if (
                    self.inputs[i] == c.toPort
                    or self.inputs[i] == c.fromPort
                    or self.outputs[i] == c.toPort
                    or self.outputs[i] == c.fromPort
            ):
                return i

    def addTree(self):
        """
        When a blockitem is added to the main window.
        A file explorer for that item is added to the right of the main window by calling this method
        """
        self.logger.debug(self.parent.parent())
        pathName = self.displayName
        if self.parent.parent().projectPath == "":
            self.path = self.parent.parent().projectFolder
        else:
            self.path = self.parent.parent().projectPath
        self.path = os.path.join(self.path, "ddck")
        self.path = os.path.join(self.path, pathName)
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.model = MyQFileSystemModel()
        self.model.setRootPath(self.path)
        self.model.setName(self.displayName)
        self.tree = MyQTreeView(self.model, self)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(self.path))
        self.tree.setObjectName("%sTree" % self.displayName)
        for i in range(1, self.model.columnCount() - 1):
            self.tree.hideColumn(i)
        self.tree.setMinimumHeight(200)
        self.tree.setSortingEnabled(True)
        self.parent.parent().splitter.addWidget(self.tree)

    def hasDdckPlaceHolders(self):
        return True

    def deleteBlock(self):
        """
        Overridden method to also delete folder
        """
        self.logger.debug("Block " + str(self) + " is deleting itself (" + self.displayName + ")")
        self.deleteConns()
        self.parent.parent().trnsysObj.remove(self)
        self.logger.debug("deleting block " + str(self) + self.displayName)
        self.parent.scene().removeItem(self)
        widgetToRemove = self.parent.parent().findChild(QTreeView, self.displayName + "Tree")
        shutil.rmtree(self.path)
        self.deleteLoadedFile()
        try:
            widgetToRemove.hide()
        except AttributeError:
            self.logger.debug("Widget doesnt exist!")
        else:
            self.logger.debug("Deleted widget")
        del self

    def setDisplayName(self, newName):
        """
        Overridden method to also change folder name
        """
        self.displayName = newName
        self.label.setPlainText(newName)
        self.model.setName(self.displayName)
        self.tree.setObjectName("%sTree" % self.displayName)
        self.logger.debug(os.path.dirname(self.path))
        destPath = os.path.join(os.path.split(self.path)[0], self.displayName)
        if os.path.exists(self.path):
            os.rename(self.path, destPath)
            self.path = destPath
            self.logger.debug(self.path)



from __future__ import annotations

import typing as _tp

import PyQt5.QtWidgets as _qtw

import trnsysGUI.common as _com

import trnsysGUI.BlockItem as _bi
import trnsysGUI.PortItemBase as _pi
import trnsysGUI.connection.connectionBase as _cb

if _tp.TYPE_CHECKING:
    import trnsysGUI.diagram.Editor as _ed


class DeleteBlockCommand(_qtw.QUndoCommand):
    def __init__(self, blockItem: _bi.BlockItem, editor: _ed.Editor):  # type: ignore[name-defined]
        description = f"Delete {blockItem.name}"
        super().__init__(description)

        self._blockItem = blockItem
        self._editor = editor

        self._createChildDeleteConnectionUndoCommands(blockItem)

    def _createChildDeleteConnectionUndoCommands(self, blockItem):
        self._createChildDeleteConnectionUndoCommandsForPorts(blockItem.inputs)
        self._createChildDeleteConnectionUndoCommandsForPorts(blockItem.outputs)

    def _createChildDeleteConnectionUndoCommandsForPorts(
        self, ports: _tp.Sequence[_pi.PortItemBase]  # type: ignore[name-defined]
    ) -> None:
        connections: _tp.Sequence[_cb.ConnectionBase] = [  # type: ignore[name-defined]
            connection for p in ports if (connection := _com.getSingleOrNone(p.connectionList)) is not None
        ]

        for connection in connections:
            connection.createDeleteUndoCommand(self)

    def redo(self):
        super().redo()

        self._blockItem.deleteBlock()

    def undo(self):
        self._editor.trnsysObj.append(self._blockItem)
        self._editor.diagramScene.addItem(self._blockItem)
        self._blockItem.addTree()

        super().undo()



import typing as _tp

import trnsysGUI.connection.singlePipeConnection as _spc
import trnsysGUI.singlePipePortItem as _spi
from . import _loopWideDefaults as _lwd
from . import search
from . import model as _model


def createLoops(
    connections: _tp.Sequence[_spc.SinglePipeConnection], fluid: _model.Fluid  # type: ignore[name-defined]
) -> _model.HydraulicLoops:
    loops = _model.HydraulicLoops([])

    todo = connections
    while todo:
        nextConnection = todo[0]
        assert isinstance(nextConnection.fromPort, _spi.SinglePipePortItem)

        reachableConnections = [*search.getReachableConnections(nextConnection.fromPort)]
        todo = [c for c in todo if c not in reachableConnections]

        name = loops.generateName()

        useLoopWideDefaults = True
        _lwd.resetConnectionPropertiesToLoopWideDefaults(reachableConnections, name.value)
        loop = _model.HydraulicLoop(name, fluid, useLoopWideDefaults, reachableConnections)

        loops.addLoop(loop)

    return loops



import enum as _enum


class Side(_enum.Enum):
    LEFT = "left"
    RIGHT = "right"

    @staticmethod
    def createFromSideNr(sideNr: int) -> "Side":
        if sideNr == 2:
            return Side.RIGHT

        if sideNr == 0:
            return Side.LEFT

        raise ValueError(f"Unknown side number: {sideNr}")

    def toSideNr(self):
        if self == self.LEFT:
            return 0

        if self == self.RIGHT:
            return 2

        raise ValueError(f"Cannot convert {self} to side nr.")

    def formatDdck(self):
        if self == self.LEFT:
            return "Left"

        if self == self.RIGHT:
            return "Right"

        raise ValueError(f"Cannot convert {self} to side nr.")

    @property
    def isLeft(self) -> bool:
        return self == self.LEFT



import trnsysGUI.massFlowSolver.networkModel as _mfn


def getEquation(outputTemp: str, massFlowRate: str, posFlowInputTemp: str, negFlowInputTemp: str) -> str:
    coldEquation = f"{outputTemp} = GE({massFlowRate}, 0)*{posFlowInputTemp} + LE({massFlowRate}, 0)*{negFlowInputTemp}"
    return coldEquation


def getMfrName(pipe: _mfn.Pipe) -> str:
    outputVariable = pipe.getOutputVariables()[0]
    assert outputVariable
    mfrName = outputVariable.name
    return mfrName



def getHeatCapacityName(loopName: str) -> str:
    return f"L{loopName}Cp"


def getDensityName(loopName: str) -> str:
    return f"L{loopName}Rho"


def getDefaultDiameterName(loopName: str) -> str:
    return f"{loopName}Dia"


def getDefaultUValueName(loopName: str) -> str:
    return f"{loopName}UVal"


def getDefaultLengthName(loopName: str) -> str:
    return f"{loopName}Len"


def getNumberOfPipesName(loopName: str) -> str:
    return f"{loopName}NPipes"



from __future__ import annotations

import typing as _tp

from PyQt5 import QtWidgets as _qtw

import pytrnsys.utils.result as _res
import trnsysGUI.connection.singlePipeConnection as _spc  # pylint: disable=cyclic-import
import trnsysGUI.hydraulicLoops.merge as _hlmerge  # pylint: disable=cyclic-import
import trnsysGUI.hydraulicLoops.model as _hlmodel
import trnsysGUI.hydraulicLoops.split as _hlsplit  # pylint: disable=cyclic-import


class DeleteSinglePipeConnectionCommand(_qtw.QUndoCommand):  # pylint: disable=too-many-instance-attributes
    def __init__(
            self,
            connection: _spc.SinglePipeConnection,
            hydraulicLoops: _hlmodel.HydraulicLoops,
            fluids: _tp.Sequence[_hlmodel.Fluid],
            defaultFluid: _hlmodel.Fluid,
            parentCommand: _tp.Optional[_qtw.QUndoCommand]
    ):
        super().__init__("Delete single pipe connection", parentCommand)
        self._connection: _tp.Optional[_spc.SinglePipeConnection] = connection
        self._hydraulicLoops = hydraulicLoops
        self._fluids = fluids
        self._defaultFluid = defaultFluid

        self._fromPort = connection.fromPort
        self._toPort = connection.toPort
        self._connectionParent = self._connection.parent

        self._splitSummary: _tp.Optional[_hlsplit.SplitSummary] = None

    def redo(self) -> None:
        assert self._connection

        splitLoopsSummary = self._splitSummary.after if self._splitSummary else None  # pylint: disable=no-member

        cancellable = _hlsplit.split(self._connection, self._hydraulicLoops, self._fluids, splitLoopsSummary)
        if cancellable == "cancelled":
            self.setObsolete(True)
            return
        self._splitSummary = cancellable

        self._connection.deleteConn()
        self._connection = None

    def undo(self) -> None:
        self._connection = _spc.SinglePipeConnection(  # type: ignore[attr-defined]
            self._fromPort, self._toPort, self._connectionParent
        )

        mergedLoopSummary = self._splitSummary.before if self._splitSummary else None  # pylint: disable=no-member

        cancellable = _hlmerge.merge(
            self._connection, self._hydraulicLoops, self._fluids, self._defaultFluid, mergedLoopSummary
        )
        assert cancellable != "cancelled" and not _res.isError(cancellable)



from __future__ import annotations

__all__ = ["resetConnectionPropertiesToLoopWideDefaults"]

import typing as _tp

import trnsysGUI.connection.values as _values

from . import names as _names


if _tp.TYPE_CHECKING:
    import trnsysGUI.connection.singlePipeConnection as _spc


def resetConnectionPropertiesToLoopWideDefaults(
    connections: _tp.Sequence[_spc.SinglePipeConnection], loopName: str  # type: ignore[name-defined]
) -> None:
    for connection in connections:
        connection.lengthInM = _values.Variable(_names.getDefaultLengthName(loopName))
        connection.diameterInCm = _values.Variable(_names.getDefaultDiameterName(loopName))
        connection.uValueInWPerM2K = _values.Variable(_names.getDefaultUValueName(loopName))



import dataclasses as _dc
import typing as _tp

import trnsysGUI.PortItemBase as _pi
from trnsysGUI.massFlowSolver import networkModel as _mfn


@_dc.dataclass
class PortItemAndAdjacentRealNode:
    portItem: _mfn.PortItem
    realNode: _mfn.RealNodeBase


@_dc.dataclass
class InternalPiping:
    openLoopsStartingNodes: _tp.Sequence[_mfn.RealNodeBase]
    modelPortItemsToGraphicalPortItem: _tp.Mapping[_mfn.PortItem, _pi.PortItemBase]  # type: ignore[name-defined]

    def getPortItemsAndAdjacentRealNodeForGraphicalPortItem(
        self, graphicalPortItem: _pi.PortItemBase  # type: ignore[name-defined]
    ) -> _tp.Sequence[PortItemAndAdjacentRealNode]:
        if graphicalPortItem not in self.modelPortItemsToGraphicalPortItem.values():
            raise ValueError("The graphical port item is not part of this internal piping.")

        portItemsAndAdjacentRealNode = []
        for startingNode in self.openLoopsStartingNodes:
            realNodesAndPortItems = _mfn.getConnectedRealNodesAndPortItems(startingNode)
            for realNode in realNodesAndPortItems.realNodes:
                for candidatePortItem in [n for n in realNode.getNeighbours() if isinstance(n, _mfn.PortItem)]:
                    candidateGraphicalPortItem = self.modelPortItemsToGraphicalPortItem[candidatePortItem]
                    if candidateGraphicalPortItem == graphicalPortItem:
                        portItem = candidatePortItem
                        portItemAndAdjacentRealNode = PortItemAndAdjacentRealNode(portItem, realNode)

                        portItemsAndAdjacentRealNode.append(portItemAndAdjacentRealNode)
        return portItemsAndAdjacentRealNode

    def getAllRealNodes(self) -> _tp.Sequence[_mfn.RealNodeBase]:
        allRealNodes: _tp.List[_mfn.RealNodeBase] = []
        for startingNode in self.openLoopsStartingNodes:
            visitedNodes = {startingNode}
            stack = [startingNode]
            while stack:
                currentNode = stack.pop()
                unvisitedNodes = [
                    n for n in currentNode.getNeighbours() if n not in visitedNodes and isinstance(n, _mfn.RealNodeBase)
                ]
                visitedNodes.update(unvisitedNodes)
                stack.extend(unvisitedNodes)

            allRealNodes.extend(visitedNodes)

        return allRealNodes


@_dc.dataclass
class _OpenLoop:
    realNodes: _tp.Sequence[_mfn.RealNodeBase]


class MassFlowNetworkContributorMixin:
    def getInternalPiping(self) -> InternalPiping:
        raise NotImplementedError()

    def exportInputsFlowSolver(self):
        openLoops, _ = self._getOpenLoopsAndNodeToIndices()

        allInputVariables = []
        for openLoop in openLoops:
            inputVariables = [n.getInputVariable() for n in openLoop.realNodes]
            allInputVariables.extend(inputVariables)

        line = ""
        for inputVariable in allInputVariables:
            if not inputVariable:
                line += f"{_mfn.InputVariable.UNDEFINED_INPUT_VARIABLE_VALUE} "
            else:
                line += f"{inputVariable.name} "

        return line, len(allInputVariables)

    def exportOutputsFlowSolver(self, prefix, abc, equationNumber, simulationUnit):  # pylint: disable=unused-argument
        openLoops, _ = self._getOpenLoopsAndNodeToIndices()

        realNodes = [n for l in openLoops for n in l.realNodes]

        lines = self._getOutputLines(realNodes, equationNumber, simulationUnit)

        joinedLines = "\n".join(lines) + "\n"
        nLinesGenerated = len(lines)
        nextEquationNumber = equationNumber + len(realNodes) * _mfn.MAX_N_OUTPUT_VARIABLES_PER_NODE

        return joinedLines, nextEquationNumber, nLinesGenerated

    @staticmethod
    def _getOutputLines(realNodes: _tp.Sequence[_mfn.RealNodeBase], equationNumber, simulationUnit):
        lines = []
        for nodeIndex, node in enumerate(realNodes):
            outputVariables = node.getOutputVariables()
            for variableIndex, outputVariable in enumerate(outputVariables):
                if not outputVariable:
                    continue

                index = equationNumber + nodeIndex * _mfn.MAX_N_OUTPUT_VARIABLES_PER_NODE + variableIndex
                line = f"{outputVariable.name}=[{simulationUnit},{index}]"
                lines.append(line)
        return lines

    def _getOpenLoopsAndNodeToIndices(self) -> _tp.Tuple[_tp.Sequence[_OpenLoop], _tp.Mapping[_mfn.NodeBase, int]]:
        internalPiping = self.getInternalPiping()
        allNodesToIndices: _tp.Dict[_mfn.NodeBase, int] = {}
        openLoops = []
        for startingNode in internalPiping.openLoopsStartingNodes:
            realNodesAndPortItems = _mfn.getConnectedRealNodesAndPortItems(startingNode)
            realNodes = realNodesAndPortItems.realNodes
            portItems = realNodesAndPortItems.portItems

            openLoops.append(_OpenLoop(realNodes))

            portItemsToIndices = {}
            for portItem in portItems:
                graphicalPortItem = internalPiping.modelPortItemsToGraphicalPortItem[portItem]
                connectedRealNode = graphicalPortItem.getConnectedRealNode(portItem, self)
                if not connectedRealNode:
                    raise AssertionError("Hydraulics not connected.")
                portItemsToIndices[portItem] = connectedRealNode.trnsysId
            realNodesToIndices = {n: n.trnsysId for n in realNodes}

            # Jump through some hoops to make `mypy` type check this
            # (see https://github.com/python/mypy/issues/1114)
            nodesAndIndex = [
                *list(portItemsToIndices.items()),
                *list(realNodesToIndices.items()),
            ]
            nodesToIndices = dict(nodesAndIndex)

            allNodesToIndices |= nodesToIndices
        return openLoops, allNodesToIndices



from __future__ import annotations

import dataclasses as _dc
import typing as _tp

from . import _helpers
from . import _loopWideDefaults as _lwd
from . import search
from . import common as _common
from . import model as _model
from ._dialogs.split import dialog as _dialog

if _tp.TYPE_CHECKING:
    import trnsysGUI.connection.singlePipeConnection as _spc


@_dc.dataclass
class SplitSummary:
    before: _common.MergedLoopSummary
    after: _common.SplitLoopsSummary

    @staticmethod
    def fromLoops(
        mergedLoop: _model.HydraulicLoop, fromLoop: _model.HydraulicLoop, toLoop: _model.HydraulicLoop
    ) -> "SplitSummary":
        mergedLoopSummary = _common.MergedLoopSummary.fromLoop(mergedLoop)
        splitLoopsSummary = _common.SplitLoopsSummary.fromLoops(fromLoop, toLoop)
        return SplitSummary(mergedLoopSummary, splitLoopsSummary)


def split(
    connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
    hydraulicLoops: _model.HydraulicLoops,
    fluids: _tp.Sequence[_model.Fluid],
    splitLoopsSummary: _tp.Optional[_common.SplitLoopsSummary],
) -> _common.Cancellable[_tp.Optional[SplitSummary]]:
    splitter = _Splitter(hydraulicLoops, fluids)

    return splitter.split(connection, splitLoopsSummary)


class _Splitter:
    def __init__(self, hydraulicLoops: _model.HydraulicLoops, fluids: _tp.Sequence[_model.Fluid]) -> None:
        self._hydraulicLoops = hydraulicLoops
        self._fluids = fluids

    def split(
        self,
        connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
        splitLoopsSummary: _tp.Optional[_common.SplitLoopsSummary] = None,
    ) -> _common.Cancellable[_tp.Optional[SplitSummary]]:
        fromPort, toPort = _helpers.getFromAndToPort(connection)

        fromConnections = search.getReachableConnections(fromPort, ignoreConnections={connection})
        toConnections = search.getReachableConnections(toPort, ignoreConnections={connection})

        loop = self._hydraulicLoops.getLoopForExistingConnection(connection)

        otherConnections = {*loop.connections} - {connection}
        isOnlyConnection = not otherConnections
        if isOnlyConnection:
            self._removeLoop(loop)
            return None

        isConnectionRedundant = (
            fromConnections == otherConnections or toConnections == otherConnections  # pylint: disable=consider-using-in
        )
        if isConnectionRedundant:
            assert (not fromConnections or fromConnections == otherConnections) and (
                not toConnections or toConnections == otherConnections
            )

            self._removeConnection(connection, loop)
            return None

        return self._splitLoop(loop, connection, splitLoopsSummary)

    @staticmethod
    def _removeConnection(connection, loop):
        loop.removeConnection(connection)

    def _removeLoop(self, loop: _model.HydraulicLoop) -> None:
        self._hydraulicLoops.removeLoop(loop)

    def _splitLoop(
        self,
        loop: _model.HydraulicLoop,
        connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
        splitLoopsSummary: _tp.Optional[_common.SplitLoopsSummary],
    ) -> _common.Cancellable[SplitSummary]:
        fromPort, toPort = _helpers.getFromAndToPort(connection)

        fromConnections = search.getReachableConnections(fromPort, ignoreConnections={connection})
        toConnections = search.getReachableConnections(toPort, ignoreConnections={connection})

        if not splitLoopsSummary:
            cancellable = self._createSplitLoopsSummary(loop, fromConnections, connection, toConnections)
            if cancellable == "cancelled":
                return "cancelled"
            splitLoopsSummary = cancellable

        self._hydraulicLoops.removeLoop(loop)

        fromLoopName = splitLoopsSummary.fromLoop.name
        toLoopName = splitLoopsSummary.toLoop.name

        useLoopWideDefaults = loop.useLoopWideDefaults
        if useLoopWideDefaults:
            _lwd.resetConnectionPropertiesToLoopWideDefaults([*fromConnections], fromLoopName.value)
            _lwd.resetConnectionPropertiesToLoopWideDefaults([*toConnections], toLoopName.value)

        fromLoop = _model.HydraulicLoop(
            fromLoopName, splitLoopsSummary.fromLoop.fluid, useLoopWideDefaults, [*fromConnections]
        )

        toLoop = _model.HydraulicLoop(
            toLoopName, splitLoopsSummary.toLoop.fluid, useLoopWideDefaults, [*toConnections]
        )

        self._hydraulicLoops.addLoop(fromLoop)
        self._hydraulicLoops.addLoop(toLoop)

        return SplitSummary.fromLoops(loop, fromLoop, toLoop)

    def _createSplitLoopsSummary(
        self,
        loop: _model.HydraulicLoop,
        fromConnections: _tp.Set[_spc.SinglePipeConnection],  # type: ignore[name-defined]
        connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
        toConnections: _tp.Set[_spc.SinglePipeConnection],  # type: ignore[name-defined]
    ) -> _common.Cancellable[_common.SplitLoopsSummary]:
        if not loop.name.isUserDefined:
            return self._createSplitLoopsSummaryForAutomaticLoop(loop)

        return self._askUserForSplitLoopsSummary(loop, fromConnections, connection, toConnections)

    def _createSplitLoopsSummaryForAutomaticLoop(self, loop: _model.HydraulicLoop) -> _common.SplitLoopsSummary:
        fromLoopSummary = _common.LoopSummary(loop.name, loop.fluid)

        toLoopName = self._hydraulicLoops.generateName()
        toLoopSummary = _common.LoopSummary(toLoopName, loop.fluid)

        return _common.SplitLoopsSummary(fromLoopSummary, toLoopSummary)

    def _askUserForSplitLoopsSummary(
        self,
        loop: _model.HydraulicLoop,
        fromConnections: _tp.Set[_spc.SinglePipeConnection],  # type: ignore[name-defined]
        connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
        toConnections: _tp.Set[_spc.SinglePipeConnection],  # type: ignore[name-defined]
    ) -> _common.Cancellable[_common.SplitLoopsSummary]:
        occupiedNames = {l.name.value for l in self._hydraulicLoops.hydraulicLoops} - {loop.name.value}

        setLoop1Selected = self._createSetConnectionsSelectedCallback(fromConnections)
        setLoop2Selected = self._createSetConnectionsSelectedCallback(toConnections)

        connection.deselectConnection()

        cancellable = _dialog.SplitLoopDialog.showDialogAndGetResult(
            loop,
            occupiedNames,
            self._fluids,
            setLoop1Selected,
            setLoop2Selected,
        )

        return cancellable

    @staticmethod
    def _createSetConnectionsSelectedCallback(
        connections: _tp.Set[_spc.SinglePipeConnection],  # type: ignore[name-defined]
    ) -> _tp.Callable[[bool], None]:
        def callback(isSelected: bool) -> None:
            _common.setConnectionsSelected(connections, isSelected)

        return callback



import typing as _tp

import PyQt5.QtWidgets as _qtw
import PyQt5.QtGui as _qtg
import PyQt5.QtCore as _qtc

from trnsysGUI.hydraulicLoops import common as _common

try:
    from . import _UI_dialog_generated as _uigen
except ImportError as importError:
    raise AssertionError(  # pylint: disable=duplicate-code  # 2
        "Could not find the generated Python code for a .ui or .qrc file. Please run the "
        "`dev-tools\\generateGuiClassesFromQtCreatorStudioUiFiles.py' script from your "
        "`pytrnsys_gui` directory."
    ) from importError

from trnsysGUI.hydraulicLoops import model as _model


class MergeLoopsDialog(_qtw.QDialog, _uigen.Ui_mergeLoopsDialog):
    def __init__(
        self,
        loop1: _model.HydraulicLoop,
        loop2: _model.HydraulicLoop,
        occupiedNames: _tp.Set[str],
        fluids: _tp.Sequence[_model.Fluid],
    ) -> None:
        super().__init__()
        self.setupUi(self)

        self._loop1 = loop1
        self._loop2 = loop2
        self._occupiedNames = occupiedNames
        self._fluids = fluids

        self.mergeSummary: _common.Cancellable[_common.MergedLoopSummary] = "cancelled"

        needsDifferingFluidsWarning = self._loop1.fluid != self._loop2.fluid
        self.differingFluidsWarningWidget.setVisible(needsDifferingFluidsWarning)
        self.differingFluidsWarningLabel.setVisible(needsDifferingFluidsWarning)

        self._configureNameComboBox()
        self._configureFluidComboBox()
        self._configureApplyButton()
        self._configureAbortButton()

        self.nameComboBox.setCurrentIndex(0)
        self.fluidComboBox.setCurrentIndex(0)

    def _configureApplyButton(self):
        def onApply() -> None:
            nameValue = self.nameComboBox.currentText()
            isUserDefined = (
                    self._loop1.name.isUserDefined
                    or self._loop2.name.isUserDefined
                    or nameValue not in [self._loop1.name and self._loop2.name]
            )
            name = _model.UserDefinedName(nameValue) if isUserDefined else _model.AutomaticallyGeneratedName(nameValue)

            fluid = self.fluidComboBox.currentData()

            self.mergeSummary = _common.MergedLoopSummary(name, fluid)

            self.close()

        self.applyButton.clicked.connect(onApply)

    def _configureAbortButton(self):
        def onAbort() -> None:
            self.mergeSummary = "cancelled"
            self.close()

        self.abortButton.clicked.connect(onAbort)

    def _configureNameComboBox(self) -> None:
        def onCurrentTextChanged(newName: str) -> None:
            self._setWarningsAndApplyEnabledForValues(newName, self.fluidComboBox.currentData())

        self.nameComboBox.currentTextChanged.connect(onCurrentTextChanged)

        self.nameComboBox.addItem("")
        self.nameComboBox.addItem(self._loop1.name.value)
        self.nameComboBox.addItem(self._loop2.name.value)

    def _configureFluidComboBox(self) -> None:
        def onCurrentIndexChanged(newIndex: int) -> None:
            name = self.nameComboBox.currentText()
            _fluid = self.fluidComboBox.itemData(newIndex) if newIndex >= 0 else None
            self._setWarningsAndApplyEnabledForValues(name, _fluid)

        self.fluidComboBox.currentIndexChanged.connect(onCurrentIndexChanged)

        fluid1 = self._loop1.fluid
        fluid2 = self._loop2.fluid
        areFluidsSame = fluid1 == fluid2
        if areFluidsSame:
            fluid = fluid1
            self._addLoopFluidItemToComboBox(fluid, [self._loop1.name.value, self._loop2.name.value])
        else:
            self.fluidComboBox.addItem("", None)
            self._addLoopFluidItemToComboBox(fluid1, [self._loop1.name.value])
            self._addLoopFluidItemToComboBox(fluid2, [self._loop2.name.value])

        otherFluids = {*self._fluids} - {self._loop1.fluid, self._loop2.fluid}
        sortedOtherFluids = sorted(otherFluids, key=lambda f: f.name)
        for fluid in sortedOtherFluids:
            self.fluidComboBox.addItem(fluid.name, fluid)

    def _setWarningsAndApplyEnabledForValues(self, name: str, fluid: _tp.Optional[_model.Fluid]) -> None:
        isNameTaken = name in self._occupiedNames
        isNameEmpty = not name
        if isNameEmpty:
            self.invalidNameWarningWidget.setToolTip("You must specify a name")
        elif isNameTaken:
            self.invalidNameWarningWidget.setToolTip("This name is already in use")

        isNameValid = not isNameEmpty and not isNameTaken
        self.invalidNameWarningWidget.setVisible(not isNameValid)

        isAFluidSelected = bool(fluid)
        self.applyButton.setEnabled(isAFluidSelected)
        self.noFluidSelectedWarningWidget.setVisible(not isAFluidSelected)

        isApplyEnabled = isNameValid and isAFluidSelected
        self.applyButton.setEnabled(isApplyEnabled)

    def _addLoopFluidItemToComboBox(self, fluid: _model.Fluid, loopNamesWithFluid: _tp.Sequence[str]) -> None:
        sortedNames = sorted(loopNamesWithFluid)
        text = f"{fluid.name} ({', '.join(sortedNames)})"
        self.fluidComboBox.addItem(text, fluid)
        boldFont = _qtg.QFont()
        boldFont.setBold(True)
        itemIndex = self.fluidComboBox.count() - 1
        self.fluidComboBox.setItemData(itemIndex, boldFont, _qtc.Qt.FontRole)

    @staticmethod
    def showDialogAndGetResult(
        loop1: _model.HydraulicLoop,
        loop2: _model.HydraulicLoop,
        occupiedNames: _tp.Set[str],
        fluids: _tp.Sequence[_model.Fluid],

    ) -> _common.Cancellable[_common.MergedLoopSummary]:
        dialog = MergeLoopsDialog(loop1, loop2, occupiedNames, fluids)
        dialog.exec()
        return dialog.mergeSummary



from __future__ import annotations

import dataclasses as _dc
import typing as _tp

import pytrnsys.utils.result as _res
import trnsysGUI.common as _com
import trnsysGUI.singlePipePortItem as _spi
from . import _loopWideDefaults as _lwd
from . import search
from . import common as _lcom
from . import model as _model, _helpers
from ._dialogs.merge import dialog as _md

if _tp.TYPE_CHECKING:
    from trnsysGUI.connection import singlePipeConnection as _spc


@_dc.dataclass
class MergeSummary:
    before: _lcom.SplitLoopsSummary
    after: _lcom.MergedLoopSummary

    @staticmethod
    def fromLoops(
        fromLoop: _model.HydraulicLoop,
        toLoop: _model.HydraulicLoop,
        mergedLoop: _model.HydraulicLoop,
    ) -> "MergeSummary":
        splitLoopsSummary = _lcom.SplitLoopsSummary.fromLoops(fromLoop, toLoop)
        mergedLoopSummary = _lcom.MergedLoopSummary.fromLoop(mergedLoop)
        mergeSummary = MergeSummary(splitLoopsSummary, mergedLoopSummary)
        return mergeSummary


def merge(
    connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
    hydraulicLoops: _model.HydraulicLoops,
    fluids: _tp.Sequence[_model.Fluid],
    defaultFluid: _model.Fluid,
    mergedLoopSummary: _tp.Optional[_lcom.MergedLoopSummary] = None,
) -> _lcom.Cancellable[_res.Result[_tp.Optional[MergeSummary]]]:
    merger = _Merger(hydraulicLoops, fluids, defaultFluid)

    return merger.merge(connection, mergedLoopSummary)


class _Merger:
    def __init__(
        self, hydraulicLoops: _model.HydraulicLoops, fluids: _tp.Sequence[_model.Fluid], defaultFluid: _model.Fluid
    ) -> None:
        self._hydraulicLoops = hydraulicLoops
        self._fluids = fluids
        self._defaultFluid = defaultFluid

    def merge(
        self,
        connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
        mergedLoopSummary: _tp.Optional[_lcom.MergedLoopSummary],
    ) -> _lcom.Cancellable[_res.Result[_tp.Optional[MergeSummary]]]:
        fromLoop, toLoop = getFromAndToLoopForNewlyCreatedConnection(self._hydraulicLoops, connection)

        if not fromLoop and not toLoop:
            self._createLoop(connection)
            return None

        if fromLoop and not toLoop:
            self._addConnection(connection, fromLoop)
            return None

        if not fromLoop and toLoop:
            self._addConnection(connection, toLoop)
            return None

        if fromLoop == toLoop:
            loop = fromLoop
            self._addConnection(connection, loop)
            return None

        # This line is needed to help mypy along
        assert fromLoop and toLoop

        return self._mergeLoops(fromLoop, toLoop, connection, mergedLoopSummary)

    @staticmethod
    def _addConnection(connection, loop):
        loop.addConnection(connection)

    def _createLoop(self, connection: _spc.SinglePipeConnection) -> None:  # type: ignore[name-defined]
        name = self._hydraulicLoops.generateName()
        connections = [connection]

        useLoopWideDefaults = True
        loop = _model.HydraulicLoop(name, self._defaultFluid, useLoopWideDefaults, connections)

        self._hydraulicLoops.addLoop(loop)

    def _mergeLoops(
        self,
        fromLoop: _model.HydraulicLoop,
        toLoop: _model.HydraulicLoop,
        connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
        mergedLoopSummary: _tp.Optional[_lcom.MergedLoopSummary],
    ) -> _lcom.Cancellable[_res.Result[MergeSummary]]:
        if fromLoop.useLoopWideDefaults != toLoop.useLoopWideDefaults:
            return _res.Error(
                "Cannot merge two loops with different settings for using loop-wide defaults. "
                "Change the loops so that their settings match and try again."
            )
        mergedUseLoopWideDefaults = fromLoop.useLoopWideDefaults

        connections = [*fromLoop.connections, connection, *toLoop.connections]
        _lcom.setConnectionsSelected(connections, True)

        if not mergedLoopSummary:
            mergedLoopSummary = self._askUserForMergedLoopSummaryOrNone(fromLoop, toLoop)

        _lcom.setConnectionsSelected(connections, False)

        if not mergedLoopSummary:
            return "cancelled"

        mergedConnections = [*fromLoop.connections, *toLoop.connections, connection]
        mergedConnections.sort(key=lambda c: c.displayName)

        mergedName = mergedLoopSummary.name

        if mergedUseLoopWideDefaults:
            _lwd.resetConnectionPropertiesToLoopWideDefaults(mergedConnections, mergedName.value)

        mergedLoop = _model.HydraulicLoop(
            mergedName, mergedLoopSummary.fluid, mergedUseLoopWideDefaults, mergedConnections
        )

        self._hydraulicLoops.removeLoop(fromLoop)
        self._hydraulicLoops.removeLoop(toLoop)
        self._hydraulicLoops.addLoop(mergedLoop)

        return MergeSummary.fromLoops(fromLoop, toLoop, mergedLoop)

    def _askUserForMergedLoopSummaryOrNone(
        self, loop1: _model.HydraulicLoop, loop2: _model.HydraulicLoop
    ) -> _tp.Optional[_lcom.MergedLoopSummary]:
        mergedName = self._getMergedNameOrNone(loop1.name, loop2.name)
        areSameFluids = loop1.fluid == loop2.fluid

        if areSameFluids and mergedName:
            mergedFluid = loop1.fluid
            return _lcom.MergedLoopSummary(mergedName, mergedFluid)

        allNames = {l.name.value for l in self._hydraulicLoops.hydraulicLoops}
        occupiedNames = allNames - {loop1.name.value, loop2.name.value}

        cancellable = _md.MergeLoopsDialog.showDialogAndGetResult(loop1, loop2, occupiedNames, self._fluids)
        if cancellable == "cancelled":
            return None
        mergedLoopSummary = cancellable

        return mergedLoopSummary

    def _getMergedNameOrNone(self, name1: _model.Name, name2: _model.Name) -> _tp.Optional[_model.Name]:
        if name1.isUserDefined and not name2.isUserDefined:
            return name1

        if not name1.isUserDefined and name2.isUserDefined:
            return name2

        if not name1.isUserDefined and not name2.isUserDefined:
            return self._hydraulicLoops.generateName()

        return None


def getFromAndToLoopForNewlyCreatedConnection(
    hydraulicLoops: _model.HydraulicLoops, connection: _spc.SinglePipeConnection  # type: ignore[name-defined]
) -> _tp.Tuple[_tp.Optional[_model.HydraulicLoop], _tp.Optional[_model.HydraulicLoop]]:
    fromPort, toPort = _helpers.getFromAndToPort(connection)

    fromLoop = _getLoopIgnoringConnection(fromPort, connection, hydraulicLoops)
    toLoop = _getLoopIgnoringConnection(toPort, connection, hydraulicLoops)

    return fromLoop, toLoop


def _getLoopIgnoringConnection(
    portItem: _spi.SinglePipePortItem,  # type: ignore[name-defined]
    connection: _spc.SinglePipeConnection,  # type: ignore[name-defined]
    hydraulicLoops: _model.HydraulicLoops,
) -> _tp.Optional[_model.HydraulicLoop]:
    connections = search.getReachableConnections(portItem, ignoreConnections={connection})
    if not connections:
        return None

    loops = {hydraulicLoops.getLoopForExistingConnection(c) for c in connections}
    loop = _com.getSingle(loops)
    return loop



# pylint: skip-file
# type: ignore

import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import dataclasses_jsonschema as _dcj

import trnsysGUI.blockItemModel as _bim
import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
import pytrnsys.utils.serialization as _ser
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin


class TeePiece(BlockItem, MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super(TeePiece, self).__init__(trnsysType, parent, **kwargs)

        self.w = 40
        self.h = 40

        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))

        self.changeSize()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.TEE_PIECE_SVG

    def changeSize(self):
        w = self.w
        h = self.h

        delta = 20
        # Limit the block size:
        if h < 20:
            h = 20
        if w < 40:
            w = 40

        # center label:
        rect = self.label.boundingRect()
        lw, lh = rect.width(), rect.height()
        lx = (w - lw) / 2

        deltaH = self.h / 8

        self.label.setPos(lx, h - self.flippedV * (h + h / 2))

        self.origInputsPos = [[0, delta]]
        self.origOutputsPos = [[w, delta], [delta, 0]]
        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])
        self.outputs[1].setPos(self.origOutputsPos[1][0], self.origOutputsPos[1][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.outputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4
        self.outputs[1].side = (self.rotationN + 1 - 2 * self.flippedV) % 4

        return w, h

    def getInternalPiping(self) -> InternalPiping:
        teePiece, modelPortItemsToGraphicalPortItem = self._getModelAndMapping()

        internalPiping = InternalPiping([teePiece], modelPortItemsToGraphicalPortItem)

        return internalPiping

    def _getModelAndMapping(self):
        input = _mfn.PortItem("input", _mfn.PortItemType.INPUT)
        output1 = _mfn.PortItem("straightOutput", _mfn.PortItemType.OUTPUT)
        output2 = _mfn.PortItem("orthogonalOutput", _mfn.PortItemType.OUTPUT)
        teePiece = _mfn.TeePiece(self.displayName, self.trnsysId, input, output1, output2)
        modelPortItemsToGraphicalPortItem = {input: self.inputs[0], output1: self.outputs[0], output2: self.outputs[1]}
        return teePiece, modelPortItemsToGraphicalPortItem

    def exportPipeAndTeeTypesForTemp(self, startingUnit):
        if self.isVisible():
            f = ""
            unitNumber = startingUnit
            tNr = 929  # Temperature calculation from a tee-piece

            unitText = ""
            ambientT = 20

            equationConstant = 1

            unitText += "UNIT " + str(unitNumber) + " TYPE " + str(tNr) + "\n"
            unitText += "!" + self.displayName + "\n"
            unitText += "PARAMETERS 0\n"
            unitText += "INPUTS 6\n"

            openLoops, nodesToIndices = self._getOpenLoopsAndNodeToIndices()
            assert len(openLoops) == 1
            openLoop = openLoops[0]

            assert len(openLoop.realNodes) == 1
            realNode = openLoop.realNodes[0]

            outputVariables = realNode.getOutputVariables()
            for outputVariable in outputVariables:
                if not outputVariable:
                    continue

                unitText += outputVariable.name + "\n"

            unitText += f"T{self.inputs[0].connectionList[0].displayName}\n"
            unitText += f"T{self.outputs[0].connectionList[0].displayName}\n"
            unitText += f"T{self.outputs[1].connectionList[0].displayName}\n"

            unitText += "***Initial values\n"
            unitText += 3 * "0 " + 3 * (str(ambientT) + " ") + "\n"

            unitText += "EQUATIONS 1\n"
            unitText += "T" + self.displayName + "= [" + str(unitNumber) + "," + str(equationConstant) + "]\n"

            unitNumber += 1
            f += unitText + "\n"

            return f, unitNumber
        else:
            return "", startingUnit
        
    def decode(self, i, resBlockList):
        model = TeePieceModel.from_dict(i)

        self.setDisplayName(model.BlockDisplayName)
        self.setPos(float(model.blockPosition[0]), float(model.blockPosition[1]))
        self.id = model.Id
        self.trnsysId = model.trnsysId
    
        self.inputs[0].id = model.portsIdsIn[0]
        self.outputs[0].id = model.portsIdsOut[0]
        self.outputs[1].id = model.portsIdsOut[1]

        self.updateFlipStateH(model.flippedH)
        self.updateFlipStateV(model.flippedV)
        self.rotateBlockToN(model.rotationN)

        resBlockList.append(self)
        
    def encode(self):
        portListInputs = []
        portListOutputs = []

        for inp in self.inputs:
            portListInputs.append(inp.id)
        for output in self.outputs:
            portListOutputs.append(output.id)

        blockPosition = (float(self.pos().x()), float(self.pos().y()))

        teePieceModel = TeePieceModel(
            self.name,
            self.displayName,
            blockPosition,
            self.id,
            self.trnsysId,
            portListInputs,
            portListOutputs,
            self.flippedH,
            self.flippedV,
            self.rotationN,
        )

        dictName = "Block-"

        dct = teePieceModel.to_dict()

        return dictName, dct


@_dc.dataclass
class TeePieceModel(_ser.UpgradableJsonSchemaMixin):  # pylint: disable=too-many-instance-attributes
    BlockName: str
    BlockDisplayName: str
    blockPosition: _tp.Tuple[float, float]
    Id: int
    trnsysId: int
    portsIdsIn: _tp.List[int]
    portsIdsOut: _tp.List[int]
    flippedH: bool
    flippedV: bool
    rotationN: int
    
    @classmethod
    def from_dict(
        cls,
        data: _dcj.JsonDict,
        validate=True,
        validate_enums: bool = True,
    ) -> "TeePieceModel":
        teePieceModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(TeePieceModel, teePieceModel)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,  # pylint: disable=duplicate-code
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)
        data[".__BlockDict__"] = True
        return data

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixin]:
        return _bim.BlockItemModel

    @classmethod
    def upgrade(cls, superseded: _bim.BlockItemModel) -> "TeePieceModel":  # type: ignore[override]
        assert len(superseded.portsIdsIn) == 2
        assert len(superseded.portsIdsOut) == 1

        inputPortIds = [superseded.portsIdsIn[0]]
        outputPortIds = [superseded.portsIdsIn[1], superseded.portsIdsOut[0]]
        
        return TeePieceModel(
            superseded.BlockName,
            superseded.BlockDisplayName,
            superseded.blockPosition,
            superseded.Id,
            superseded.trnsysId,
            inputPortIds,
            outputPortIds,
            superseded.flippedH,
            superseded.flippedV,
            superseded.rotationN,
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('3fff9a8a-d40e-42e2-824d-c015116d0a1d')



import dataclasses as _dc

import trnsysGUI.singlePipePortItem as _port
import trnsysGUI.storageTank.side as _sd


@_dc.dataclass
class DirectPortPair:
    trnsysId: int
    fromPort: _port.SinglePipePortItem  # type: ignore[name-defined]
    toPort: _port.SinglePipePortItem  # type: ignore[name-defined]
    relativeInputHeight: float
    relativeOutputHeight: float
    side: _sd.Side

    @property
    def relativeInputHeightPercent(self):
        return self._toPercent(self.relativeInputHeight)

    @property
    def relativeOutputHeightPercent(self):
        return self._toPercent(self.relativeOutputHeight)

    def setRelativeHeights(self,
                           relativeInputHeight: float,
                           relativeOutputHeight: float,
                           storageTankHeight: float) -> None:
        self.relativeInputHeight = relativeInputHeight
        self.relativeOutputHeight = relativeOutputHeight

        inputPosY = storageTankHeight - storageTankHeight*self.relativeInputHeight
        self.fromPort.setY(inputPosY)

        outputPosY = storageTankHeight - storageTankHeight*self.relativeOutputHeight
        self.toPort.setY(outputPosY)

    @staticmethod
    def _toPercent(relativeHeight: float):
        return int(round(relativeHeight * 100, 2))



# pylint: skip-file
# type: ignore

import typing as _tp

import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin


class WTap_main(BlockItem, MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super(WTap_main, self).__init__(trnsysType, parent, **kwargs)
        self.w = 40
        self.h = 40

        self.outputs.append(_cspi.createSinglePipePortItem("o", 0, self))

        self.changeSize()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.W_TAP_MAIN_SVG

    def changeSize(self):
        w = self.w
        h = self.h

        """ Resize block function """
        delta = 20

        # Limit the block size:
        if h < 20:
            h = 20
        if w < 40:
            w = 40
        # center label:
        rect = self.label.boundingRect()
        lw, lh = rect.width(), rect.height()
        lx = (w - lw) / 2
        self.label.setPos(lx, h)

        self.origOutputsPos = [[0, delta]]
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.outputs[0].side = (self.rotationN + 2 * self.flippedH) % 4

        return w, h

    def exportBlackBox(self):
        equation = ["T" + self.displayName + "=Tcw"]
        return "success", equation

    def exportMassFlows(self):
        resStr = "Mfr" + self.displayName + " = 1000" + "\n"
        equationNr = 1
        return resStr, equationNr

    def getInternalPiping(self) -> InternalPiping:
        outputPort = _mfn.PortItem("output", _mfn.PortItemType.OUTPUT)
        source = _mfn.Source(self.displayName, self.trnsysId, outputPort)
        return InternalPiping([source], {outputPort: self.outputs[0]})



# pylint: skip-file
# type: ignore

import json

from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.diagram import Editor as _de


class Encoder(json.JSONEncoder):
    """
    This class encodes the diagram (entire diagram or clipboard)
    obj is passed along to get some of the diagram editor porperties, for instance the id-generator
    TrnsysObj are then stored in a list of dictionaries, saved to a json file.
    Important: There is a slight naming error, since the dict containing all sub dicts (dicts for each block, connection
    and Id-generator) has the key ".__BlockDct__", although there are also other elements in there.
    """

    def default(self, obj):
        """

        Parameters
        ----------
        obj
        :type obj: trnsysGUI.DiagramEditor.DiagramEditor

        Returns
        -------

        """
        logger = obj.logger

        if isinstance(obj, _de.Editor):
            res = {}
            blockDct = {".__BlockDct__": True}

            for t in obj.trnsysObj:
                if isinstance(t, BlockItem) and not t.isVisible():
                    logger.debug("Invisible block [probably an insideBlock?]" + str(t) + str(t.displayName))
                    continue

                dictName, dct = t.encode()
                blockDct[dictName + str(t.id)] = dct

            idDict = {
                "__idDct__": True,
                "GlobalId": obj.idGen.getID(),
                "trnsysID": obj.idGen.getTrnsysID(),
                "globalConnID": obj.idGen.getConnID(),
            }
            blockDct["IDs"] = idDict

            nameDict = {"__nameDct__": True, "DiagramName": obj.diagramName, "ProjectFolder": obj.projectFolder}
            blockDct["Strings"] = nameDict

            for gi in obj.graphicalObj:
                dictName, dct = gi.encode()
                blockDct[dictName + str(gi.id)] = dct

            res["Blocks"] = blockDct

            res["hydraulicLoops"] = obj.hydraulicLoops.toJson()

            return res
        else:
            logger.debug("This is a strange object in Encoder" + type(obj))
            # return super().default(obj)



from __future__ import annotations

__all__ = ["Fluid", "HydraulicLoop", "HydraulicLoops", "Fluids"]

import abc as _abc
import itertools as _it
import typing as _tp

import trnsysGUI.common as _com

from . import _serialization as _ser

if _tp.TYPE_CHECKING:
    import trnsysGUI.connection.singlePipeConnection as _spc  # type: ignore[name-defined]

Fluid = _ser.Fluid
Variable = _ser.Variable


class Name(_abc.ABC):
    def __init__(self, value: str):
        self.value = value

    @property
    @_abc.abstractmethod
    def isUserDefined(self) -> bool:
        raise NotImplementedError()


class UserDefinedName(Name):
    @property
    def isUserDefined(self) -> bool:
        return True


class AutomaticallyGeneratedName(Name):
    @property
    def isUserDefined(self) -> bool:
        return False


class HydraulicLoop:
    def __init__(
        self,
        name: Name,
        fluid: _ser.Fluid,
        useLoopWideDefaults: bool,
        connections: _tp.Sequence[_spc.SinglePipeConnection],  # type: ignore[name-defined]
    ) -> None:
        self.name = name
        self.fluid = fluid
        self.connections: _tp.List[_spc.SinglePipeConnection] = [*connections]  # type: ignore[name-defined]
        self.useLoopWideDefaults = useLoopWideDefaults

    def addConnection(self, connection: _spc.SinglePipeConnection) -> None:  # type: ignore[name-defined]
        if self.containsConnection(connection):
            raise ValueError(f"Connection '{connection.displayName}' already part of hydraulic loop '{self.name}'.")

        self.connections.append(connection)

    def removeConnection(self, connection: _spc.SinglePipeConnection) -> None:  # type: ignore[name-defined]
        if not self.containsConnection(connection):
            raise ValueError(f"Connection '{connection.displayName}' does not belong to hydraulic loop {self.name}.")

        self.connections.remove(connection)

    def containsConnection(self, connection: _spc.SinglePipeConnection):  # type: ignore[name-defined]
        return connection in self.connections


class HydraulicLoops:
    def __init__(self, hydraulicLoops: _tp.Sequence[HydraulicLoop]) -> None:
        duplicateNames = _getDuplicates(l.name.value for l in hydraulicLoops)
        if duplicateNames:
            raise ValueError(
                f"Hydraulic loop names must be unique (the following names were duplicated: {','.join(duplicateNames)})."
            )

        self.hydraulicLoops = list(hydraulicLoops)

    @classmethod
    def createFromJson(
        cls,
        sequence: _tp.Sequence[_tp.Dict],
        connections: _tp.Sequence[_spc.SinglePipeConnection],  # type: ignore[name-defined]
        fluids: "Fluids",
    ) -> "HydraulicLoops":
        serializedLoops = [_ser.HydraulicLoop.from_dict(o) for o in sequence]

        connectionsByTrnsysId = {c.trnsysId: c for c in connections}

        loops = []
        for serializedLoop in serializedLoops:
            loop = cls._createLoop(serializedLoop, connectionsByTrnsysId, fluids)
            loops.append(loop)

        return HydraulicLoops(loops)

    @classmethod
    def _createLoop(
        cls,
        serializedLoop: _ser.HydraulicLoop,
        connectionsByTrnsysId: _tp.Mapping[int, _spc.SinglePipeConnection],  # type: ignore[name-defined]
        fluids: "Fluids",
    ) -> HydraulicLoop:
        name = cls._createName(serializedLoop)

        fluid = fluids.getFluid(serializedLoop.fluidName)
        assert fluid, f"Unknown fluid {serializedLoop.fluidName}"

        useLoopWideDefaults = serializedLoop.useLoopWideDefaults

        connections = [connectionsByTrnsysId[i] for i in serializedLoop.connectionsTrnsysId]

        loop = HydraulicLoop(name, fluid, useLoopWideDefaults, connections)
        return loop

    def toJson(self) -> _tp.Sequence[_tp.Dict]:
        result = []
        for loop in self.hydraulicLoops:
            connectionTrnsysIds = [c.trnsysId for c in loop.connections]
            serializedLoop = _ser.HydraulicLoop(
                loop.name.value, loop.name.isUserDefined, loop.fluid.name, loop.useLoopWideDefaults, connectionTrnsysIds
            )
            json = serializedLoop.to_dict()
            result.append(json)

        return result

    @staticmethod
    def _createName(serializedLoop: _ser.HydraulicLoop) -> Name:
        isUserDefinedName = serializedLoop.hasUserDefinedName
        nameValue = serializedLoop.name
        name = UserDefinedName(nameValue) if isUserDefinedName else AutomaticallyGeneratedName(nameValue)
        return name

    def getLoopForExistingConnection(
        self, connection: _spc.SinglePipeConnection  # type: ignore[name-defined]
    ) -> HydraulicLoop:
        loops = {l for l in self.hydraulicLoops if connection in l.connections}
        loop = _com.getSingle(loops)
        return loop

    def getLoop(self, name: str) -> _tp.Optional[HydraulicLoop]:
        loop = _getSingleOrNone(l for l in self.hydraulicLoops if l.name.value == name)
        return loop

    def addLoop(self, loop: HydraulicLoop) -> None:
        if self.getLoop(loop.name.value):
            raise ValueError(f"Loop with name '{loop.name.value}' already exists.")

        self.hydraulicLoops.append(loop)

    def removeLoop(self, loop: HydraulicLoop) -> None:
        self.hydraulicLoops.remove(loop)

    def clear(self):
        self.hydraulicLoops = []

    def generateName(self) -> AutomaticallyGeneratedName:
        for i in range(1, 100000):
            candidateName = f"loop{i}"
            if not self.getLoop(candidateName):
                generatedName = candidateName
                return AutomaticallyGeneratedName(generatedName)

        raise AssertionError(f"Failed to generate new name after {i} tries.")


class Fluids:
    WATER = _ser.Fluid("water", specificHeatCapacityInJPerKgK=Variable("CPWAT_SI"), densityInKgPerM3=Variable("RHOWAT"))
    BRINE = _ser.Fluid("brine", specificHeatCapacityInJPerKgK=Variable("CPBRI_SI"), densityInKgPerM3=Variable("RHOBRI"))

    def __init__(self, fluids: _tp.Sequence[_ser.Fluid]) -> None:
        duplicateNames = _getDuplicates(f.name for f in fluids)
        if duplicateNames:
            raise ValueError(
                f"Fluid names must be unique (the following names were duplicated: {','.join(duplicateNames)})."
            )

        fluidList = [*fluids]

        haveWater = any(f.name == self.WATER.name for f in fluidList)
        if not haveWater:
            fluidList.append(self.WATER)

        haveBrine = any(f.name == self.BRINE.name for f in fluidList)
        if not haveBrine:
            fluidList.append(self.BRINE)

        self.fluids = sorted(fluidList, key=lambda f: f.name)

    @staticmethod
    def createFromJson(sequence: _tp.Sequence[_tp.Dict]) -> "Fluids":
        fluids = [_ser.Fluid.from_dict(o) for o in sequence]
        return Fluids(fluids)

    def toJson(self) -> _tp.Sequence[_tp.Dict]:
        result = [f.to_dict() for f in self.fluids]
        return result

    def getFluid(self, name: str) -> _tp.Optional[_ser.Fluid]:
        fluid = _getSingleOrNone(f for f in self.fluids if f.name == name)
        return fluid


_TValue = _tp.TypeVar("_TValue", covariant=True)


class _SupportsLessThan(_tp.Protocol):
    def __lt__(self, other: _tp.Any) -> bool:
        ...


_SupportsLessThanT = _tp.TypeVar("_SupportsLessThanT", bound=_SupportsLessThan)


def _getDuplicates(values: _tp.Iterable[_SupportsLessThanT]) -> _tp.Sequence[_SupportsLessThanT]:
    sortedValues = sorted(values)
    groupedValues = _it.groupby(sortedValues)
    duplicateValues = [k for k, g in groupedValues if len(list(g)) > 1]
    return duplicateValues


def _getSingleOrNone(values: _tp.Iterable[_TValue]) -> _tp.Optional[_TValue]:
    valuesList = [*values]
    if not valuesList:
        return None

    numberOfValues = len(valuesList)
    assert numberOfValues == 1, f"Expected 0 or 1 value, but got {numberOfValues}."

    return valuesList[0]



from __future__ import annotations

import typing as _tp

import trnsysGUI.connectors.doublePipeConnectorBase as _dpcb
import trnsysGUI.doublePipePortItem as _dppi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn

from trnsysGUI import connectorsAndPipesExportHelpers as _helpers

if _tp.TYPE_CHECKING:
    import trnsysGUI.connection.doublePipeConnection as _dpc


class DoubleDoublePipeConnector(_dpcb.DoublePipeConnectorBase):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.inputs.append(_dppi.DoublePipePortItem("i", 0, self))
        self.outputs.append(_dppi.DoublePipePortItem("o", 2, self))

        self.changeSize()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.DOUBLE_DOUBLE_PIPE_CONNECTOR_SVG

    def changeSize(self):
        super().changeSize()

        self.origInputsPos = [[0, 10]]
        self.origOutputsPos = [[40, 10]]

        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        # pylint: disable=duplicate-code  # 3
        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.outputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4

    def getInternalPiping(self) -> _mfs.InternalPiping:
        coldModelPortItemsToGraphicalPortItem = {
            self._coldPipe.fromPort: self.outputs[0],
            self._coldPipe.toPort: self.inputs[0],
        }

        hotModelPortItemsToGraphicalPortItem = {
            self._hotPipe.fromPort: self.inputs[0],
            self._hotPipe.toPort: self.outputs[0],
        }

        modelPortItemsToGraphicalPortItem = coldModelPortItemsToGraphicalPortItem | hotModelPortItemsToGraphicalPortItem

        internalPiping = _mfs.InternalPiping([self._coldPipe, self._hotPipe], modelPortItemsToGraphicalPortItem)

        return internalPiping

    def exportPipeAndTeeTypesForTemp(self, startingUnit: int) -> _tp.Tuple[str, int]:
        fromConnection = self.inputs[0].getConnection()
        toConnection = self.outputs[0].getConnection()

        hotEquation = self._getEquation("Hot", self._hotPipe, fromConnection, toConnection)
        coldEquation = self._getEquation("Cold", self._coldPipe, toConnection, fromConnection)

        equations = f"""\
EQUATIONS 2
{coldEquation}
{hotEquation}
"""
        return equations, startingUnit

    def _getEquation(
        self,
        temperatureSuffix: str,
        pipe: _mfn.Pipe,
        posFlowInputConnection: _dpc.DoublePipeConnection,
        negFlowInputConnection: _dpc.DoublePipeConnection,
    ) -> str:
        hotOutputTemp = f"T{self.displayName}{temperatureSuffix}"
        hotMfr = _helpers.getMfrName(pipe)
        posFlowHotInputTemp = f"T{posFlowInputConnection.displayName}{temperatureSuffix}"
        negFlowHotInputTemp = f"T{negFlowInputConnection.displayName}{temperatureSuffix}"
        hotEquation = _helpers.getEquation(
            hotOutputTemp, hotMfr, posFlowHotInputTemp, negFlowHotInputTemp
        )
        return hotEquation



from __future__ import annotations

import typing as _tp

from trnsysGUI.hydraulicLoops import search

import trnsysGUI.singlePipePortItem as _spc

if _tp.TYPE_CHECKING:
    import trnsysGUI.BlockItem as _bi


def createSinglePipePortItem(
        name: str, side: _spc.Side, parent: _bi.BlockItem
) -> _spc.SinglePipePortItem:
    return _spc.SinglePipePortItem(name, side, parent, search.getInternallyConnectedPortItems)



from __future__ import annotations

import typing as _tp

from PyQt5.QtCore import QPoint
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QGraphicsView

from trnsysGUI.AirSourceHP import AirSourceHP  # type: ignore[attr-defined]
from trnsysGUI.BlockItem import BlockItem  # type: ignore[attr-defined]
from trnsysGUI.Boiler import Boiler  # type: ignore[attr-defined]
from trnsysGUI.Collector import Collector  # type: ignore[attr-defined]
from trnsysGUI.connectors.connector import Connector  # type: ignore[attr-defined]
from trnsysGUI.Control import Control  # type: ignore[attr-defined]
from trnsysGUI.ExternalHx import ExternalHx  # type: ignore[attr-defined]
from trnsysGUI.GenericBlock import GenericBlock  # type: ignore[attr-defined]
from trnsysGUI.Graphicaltem import GraphicalItem  # type: ignore[attr-defined]
from trnsysGUI.GroundSourceHx import GroundSourceHx  # type: ignore[attr-defined]
from trnsysGUI.HPDoubleDual import HPDoubleDual  # type: ignore[attr-defined]
from trnsysGUI.HPDual import HPDual  # type: ignore[attr-defined]
from trnsysGUI.HeatPump import HeatPump  # type: ignore[attr-defined]
from trnsysGUI.HeatPumpTwoHx import HeatPumpTwoHx  # type: ignore[attr-defined]
from trnsysGUI.IceStorage import IceStorage  # type: ignore[attr-defined]
from trnsysGUI.IceStorageTwoHx import IceStorageTwoHx  # type: ignore[attr-defined]
from trnsysGUI.MasterControl import MasterControl  # type: ignore[attr-defined]
from trnsysGUI.PV import PV  # type: ignore[attr-defined]
from trnsysGUI.PitStorage import PitStorage  # type: ignore[attr-defined]
from trnsysGUI.pump import Pump  # type: ignore[attr-defined]
from trnsysGUI.Radiator import Radiator  # type: ignore[attr-defined]
from trnsysGUI.TVentil import TVentil  # type: ignore[attr-defined]
from trnsysGUI.TeePiece import TeePiece  # type: ignore[attr-defined]
from trnsysGUI.WTap import WTap  # type: ignore[attr-defined]
from trnsysGUI.WTap_main import WTap_main  # type: ignore[attr-defined]
from trnsysGUI.crystalizer import Crystalizer
from trnsysGUI.deleteBlockCommand import DeleteBlockCommand
from trnsysGUI.connectors.doubleDoublePipeConnector import DoubleDoublePipeConnector
from trnsysGUI.doublePipeTeePiece import DoublePipeTeePiece
from trnsysGUI.geotherm import Geotherm
from trnsysGUI.connectors.singleDoublePipeConnector import SingleDoublePipeConnector
from trnsysGUI.sink import Sink
from trnsysGUI.source import Source
from trnsysGUI.sourceSink import SourceSink
from trnsysGUI.storageTank.widget import StorageTank
from trnsysGUI.water import Water

if _tp.TYPE_CHECKING:
    import trnsysGUI.diagram.Editor as _ed


class View(QGraphicsView):
    """
    Displays the items from the Scene. Here, the drag and drop from the library to the View is implemented.

    """

    def __init__(self, scene, editor: _ed.Editor) -> None:  # type: ignore[name-defined]
        super().__init__(scene, editor)

        self.logger = editor.logger
        self._editor = editor

        self.adjustSize()
        self.setRenderHint(QPainter.Antialiasing)

    def dragEnterEvent(self, event):  # pylint: disable=no-self-use
        if event.mimeData().hasFormat("component/name"):
            event.accept()

    def dragMoveEvent(self, event):  # pylint: disable=no-self-use
        if event.mimeData().hasFormat("component/name"):
            event.accept()

    def dropEvent(self, event):  # pylint: disable=too-many-branches,too-many-statements
        """Here, the dropped icons create BlockItems/GraphicalItems"""
        if event.mimeData().hasFormat("component/name"):
            componentType = str(event.mimeData().data("component/name"), encoding="utf-8")
            self.logger.debug("name is " + componentType)
            if componentType == "StorageTank":
                blockItem = StorageTank(componentType, self, displayNamePrefix="Tes")
                self.parent().showConfigStorageDlg(blockItem)
            elif componentType == "TeePiece":
                blockItem = TeePiece(componentType, self, displayNamePrefix="Tee")
            elif componentType == "DPTee":
                blockItem = DoublePipeTeePiece(componentType, self, displayNamePrefix="DTee")
            elif componentType == "SPCnr":
                blockItem = SingleDoublePipeConnector(componentType, self, displayNamePrefix="SCnr")
            elif componentType == "DPCnr":
                blockItem = DoubleDoublePipeConnector(componentType, self, displayNamePrefix="DCnr")
            elif componentType == "TVentil":
                blockItem = TVentil(componentType, self, displayNamePrefix="Val")
            elif componentType == "Pump":
                blockItem = Pump(componentType, self, displayNamePrefix="Pump")
            elif componentType == "Collector":
                blockItem = Collector(componentType, self, displayNamePrefix="Coll")
            elif componentType == "HP":
                blockItem = HeatPump(componentType, self, displayNamePrefix="HP")
            elif componentType == "IceStorage":
                blockItem = IceStorage(componentType, self, displayNamePrefix="IceS")
            elif componentType == "PitStorage":
                blockItem = PitStorage(componentType, self, displayNamePrefix="PitS")
            elif componentType == "Radiator":
                blockItem = Radiator(componentType, self, displayNamePrefix="Rad")
            elif componentType == "WTap":
                blockItem = WTap(componentType, self, displayNamePrefix="WtTp")
            elif componentType == "WTap_main":
                blockItem = WTap_main(componentType, self, displayNamePrefix="WtSp")
            elif componentType == "Connector":
                blockItem = Connector(componentType, self, displayNamePrefix="Conn")
            elif componentType == "GenericBlock":
                blockItem = GenericBlock(componentType, self, displayNamePrefix="GBlk")
                self.parent().showGenericPortPairDlg(blockItem)
            elif componentType == "HPTwoHx":
                blockItem = HeatPumpTwoHx(componentType, self, displayNamePrefix="HP")
            elif componentType == "HPDoubleDual":
                blockItem = HPDoubleDual(componentType, self, displayNamePrefix="HPDD")
            elif componentType == "HPDual":
                blockItem = HPDual(componentType, self, displayNamePrefix="HPDS")
            elif componentType == "Boiler":
                blockItem = Boiler(componentType, self, displayNamePrefix="Bolr")
            elif componentType == "AirSourceHP":
                blockItem = AirSourceHP(componentType, self, displayNamePrefix="Ashp")
            elif componentType == "PV":
                blockItem = PV(componentType, self, displayNamePrefix="PV")
            elif componentType == "GroundSourceHx":
                blockItem = GroundSourceHx(componentType, self, displayNamePrefix="Gshx")
            elif componentType == "ExternalHx":
                blockItem = ExternalHx(componentType, self, displayNamePrefix="Hx")
            elif componentType == "IceStorageTwoHx":
                blockItem = IceStorageTwoHx(componentType, self, displayNamePrefix="IceS")
            elif componentType == "GraphicalItem":
                blockItem = GraphicalItem(self)
            elif componentType == "MasterControl":
                blockItem = MasterControl(componentType, self)
            elif componentType == "Control":
                blockItem = Control(componentType, self)
            elif componentType == "Sink":
                blockItem = Sink(componentType, self, displayNamePrefix="QSnk")
            elif componentType == "Source":
                blockItem = Source(componentType, self, displayNamePrefix="QSrc")
            elif componentType == "SourceSink":
                blockItem = SourceSink(componentType, self, displayNamePrefix="QExc")
            elif componentType == "Geotherm":
                blockItem = Geotherm(componentType, self, displayNamePrefix="GeoT")
            elif componentType == "Water":
                blockItem = Water(componentType, self, displayNamePrefix="QWat")
            elif componentType == "Crystalizer":
                blockItem = Crystalizer(componentType, self, displayNamePrefix="Cryt")
            else:
                blockItem = BlockItem(componentType, self, displayNamePrefix="Blk")

            snapSize = self.parent().snapSize
            if self.parent().snapGrid:
                position = QPoint(
                    event.pos().x() - event.pos().x() % snapSize, event.pos().y() - event.pos().y() % snapSize
                )
                scenePosition = self.mapToScene(position)
            else:
                scenePosition = self.mapToScene(event.pos())

            blockItem.setPos(scenePosition)
            self.scene().addItem(blockItem)

            blockItem.oldPos = blockItem.scenePos()

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        self.parent().mouseMoveEvent(event)

    def wheelEvent(self, event):
        super().wheelEvent(event)
        if int(event.modifiers()) == 0b100000000000000000000000000:
            if event.angleDelta().y() > 0:
                self.scale(1.2, 1.2)
            else:
                self.scale(0.8, 0.8)

    def deleteBlockCom(self, blockItem):
        command = DeleteBlockCommand(blockItem, self._editor)
        self.parent().parent().undoStack.push(command)



from __future__ import annotations

import dataclasses as _dc
import typing as _tp

from . import model as _model

if _tp.TYPE_CHECKING:
    import trnsysGUI.connection.singlePipeConnection as _spc


@_dc.dataclass
class LoopSummaryBase:
    name: _model.Name
    fluid: _model.Fluid


class LoopSummary(LoopSummaryBase):
    @staticmethod
    def fromLoop(loop: _model.HydraulicLoop) -> "LoopSummary":
        return LoopSummary(loop.name, loop.fluid)


class MergedLoopSummary(LoopSummaryBase):
    @staticmethod
    def fromLoop(loop: _model.HydraulicLoop) -> "MergedLoopSummary":
        return MergedLoopSummary(loop.name, loop.fluid)


@_dc.dataclass
class SplitLoopsSummary:
    fromLoop: LoopSummary
    toLoop: LoopSummary

    @staticmethod
    def fromLoops(
        fromLoop: _model.HydraulicLoop,
        toLoop: _model.HydraulicLoop,
    ) -> "SplitLoopsSummary":
        fromLoopSummary = LoopSummary.fromLoop(fromLoop)
        toLoopSummary = LoopSummary.fromLoop(toLoop)
        return SplitLoopsSummary(fromLoopSummary, toLoopSummary)


T = _tp.TypeVar("T")  # pylint: disable=invalid-name
Cancellable = _tp.Union[_tp.Literal["cancelled"], T]


def setConnectionsSelected(
        connections: _tp.Iterable[_spc.SinglePipeConnection],  # type: ignore[name-defined]
        isSelected: bool) -> None:
    for connection in connections:
        if isSelected:
            connection.selectConnection()
        else:
            connection.deselectConnection()



import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import dataclasses_jsonschema as _dcj

import pytrnsys.utils.serialization as _ser


@_dc.dataclass
class BlockItemModelVersion0(_ser.UpgradableJsonSchemaMixinVersion0):  # pylint: disable=too-many-instance-attributes
    BlockName: str  # pylint: disable=invalid-name
    BlockDisplayName: str  # pylint: disable=invalid-name
    BlockPosition: _tp.Tuple[float, float]  # pylint: disable=invalid-name
    ID: int  # pylint: disable=invalid-name
    trnsysID: int
    PortsIDIn: _tp.List[int]  # pylint: disable=invalid-name
    PortsIDOut: _tp.List[int]  # pylint: disable=invalid-name
    FlippedH: bool  # pylint: disable=invalid-name
    FlippedV: bool  # pylint: disable=invalid-name
    RotationN: int  # pylint: disable=invalid-name
    GroupName: str  # pylint: disable=invalid-name

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("b87a3360-eaa7-48f3-9bed-d01224727cbe")


@_dc.dataclass
class BlockItemModel(_ser.UpgradableJsonSchemaMixin):  # pylint: disable=too-many-instance-attributes
    BlockName: str  # pylint: disable=invalid-name
    BlockDisplayName: str  # pylint: disable=invalid-name
    blockPosition: _tp.Tuple[float, float]
    Id: int  # pylint: disable=invalid-name
    trnsysId: int
    portsIdsIn: _tp.List[int]
    portsIdsOut: _tp.List[int]
    flippedH: bool
    flippedV: bool
    rotationN: int

    @classmethod
    def from_dict(
            cls,
            data: _dcj.JsonDict,  # pylint: disable=duplicate-code
            validate=True,
            validate_enums: bool = True,
    ) -> "BlockItemModel":
        del data[".__BlockDict__"]
        blockItemModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(BlockItemModel, blockItemModel)

    def to_dict(
            self,
            omit_none: bool = True,
            validate: bool = False,
            validate_enums: bool = True,  # pylint: disable=duplicate-code
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)
        data[".__BlockDict__"] = True
        return data

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return BlockItemModelVersion0

    @classmethod
    def upgrade(cls, superseded: BlockItemModelVersion0) -> "BlockItemModel":  # type: ignore[override]

        return BlockItemModel(
            superseded.BlockName,
            superseded.BlockDisplayName,
            superseded.BlockPosition,
            superseded.ID,
            superseded.trnsysID,
            superseded.PortsIDIn,
            superseded.PortsIDOut,
            superseded.FlippedH,
            superseded.FlippedV,
            superseded.RotationN,
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("bbc03f36-d1a1-4d97-a9c0-d212ea3a0203")



import typing as _tp

import trnsysGUI.connectors.doublePipeConnectorBase as _dpcb
import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.doublePipePortItem as _dppi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn

from trnsysGUI import connectorsAndPipesExportHelpers as _helpers


class SingleDoublePipeConnector(_dpcb.DoublePipeConnectorBase):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))
        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))
        self.outputs.append(_dppi.DoublePipePortItem("o", 2, self))

        self.changeSize()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.SINGLE_DOUBLE_PIPE_CONNECTOR_SVG

    def changeSize(self):
        super().changeSize()

        self.origInputsPos = [[0, 0], [0, 20]]
        self.origOutputsPos = [[40, 10]]

        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.inputs[1].setPos(self.origInputsPos[1][0], self.origInputsPos[1][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        # pylint: disable=duplicate-code
        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.inputs[1].side = (self.rotationN + 2 * self.flippedH) % 4
        # pylint: disable=duplicate-code
        self.outputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4

    def getInternalPiping(self) -> _mfs.InternalPiping:
        coldModelPortItemsToGraphicalPortItem = {
            self._coldPipe.fromPort: self.outputs[0],
            self._coldPipe.toPort: self.inputs[1],
        }
        hotModelPortItemsToGraphicalPortItem = {
            self._hotPipe.fromPort: self.inputs[0],
            self._hotPipe.toPort: self.outputs[0],
        }
        modelPortItemsToGraphicalPortItem = coldModelPortItemsToGraphicalPortItem | hotModelPortItemsToGraphicalPortItem
        internalPiping = _mfs.InternalPiping([self._coldPipe, self._hotPipe], modelPortItemsToGraphicalPortItem)
        return internalPiping

    def getTemperatureVariableName(self, portItem) -> str:
        internalPiping = self.getInternalPiping()

        startingNodes = internalPiping.openLoopsStartingNodes
        assert all(isinstance(n, _mfn.Pipe) for n in startingNodes)
        internalPipes = [_tp.cast(_mfn.Pipe, n) for n in startingNodes]

        graphicalPortItems = internalPiping.modelPortItemsToGraphicalPortItem

        keyPortItems = [key for key, value in graphicalPortItems.items() if value == portItem]
        assert len(keyPortItems) == 1, "portItem is not unique"
        keyPortItem = keyPortItems[0]

        for internalPipe in internalPipes:
            if keyPortItem in (internalPipe.fromNode, internalPipe.toNode):
                return "T" + internalPipe.name

        raise AssertionError("Found no internal SingleDoublePipeConnector-Connection.")

    def exportPipeAndTeeTypesForTemp(self, startingUnit: int) -> _tp.Tuple[str, int]:  # pylint: disable=too-many-locals
        doubleConnection = self.outputs[0].getConnection()

        hotInputSingleConnection = self.inputs[0].getConnection()
        coldOutputSingleConnection = self.inputs[1].getConnection()

        coldOutputTemp = f"T{self.displayName}Cold"
        coldMfr = _helpers.getMfrName(self._coldPipe)
        posFlowColdInputTemp = f"T{doubleConnection.displayName}Cold"
        negFlowColdInputTemp = f"T{coldOutputSingleConnection.displayName}"

        coldEquation = _helpers.getEquation(
            coldOutputTemp, coldMfr, posFlowColdInputTemp, negFlowColdInputTemp
        )

        hotOutputTemp = f"T{self.displayName}Hot"
        hotMfr = _helpers.getMfrName(self._hotPipe)
        posFlowHotInputTemp = f"T{hotInputSingleConnection.displayName}"
        negFlowHotInputTemp = f"T{doubleConnection.displayName}Hot"

        hotEquation = _helpers.getEquation(
            hotOutputTemp, hotMfr, posFlowHotInputTemp, negFlowHotInputTemp
        )

        equations = f"""\
EQUATIONS 2
{coldEquation}
{hotEquation}
"""
        return equations, startingUnit



# pylint: disable = invalid-name

from __future__ import annotations

import typing as _tp

from PyQt5 import QtCore
from PyQt5.QtCore import QRectF, QPointF
from PyQt5.QtGui import QColor, QBrush, QCursor, QPen
from PyQt5.QtWidgets import QGraphicsEllipseItem

if _tp.TYPE_CHECKING:
    import trnsysGUI.massFlowSolver.networkModel as _mfn
    import trnsysGUI.massFlowSolver as _mfs
    import trnsysGUI.connection.connectionBase as _cb


class PortItemBase(QGraphicsEllipseItem):  # pylint: disable = too-many-instance-attributes
    def __init__(self, name, side, parent):
        self.parent = parent

        self.logger = parent.logger

        self.name = name
        self.side = side
        self.createdAtSide = side
        self.posCallbacks = []
        self.connectionList = []
        self.id = self.parent.parent.parent().idGen.getID()  # pylint: disable = invalid-name

        self.color = "white"
        self.ashColorR = QColor(239, 57, 75)
        self.ashColorB = QColor(20, 83, 245)

        QGraphicsEllipseItem.__init__(self, QRectF(-4, -4, 7.0, 7.0), parent)

        self.innerCircle = QGraphicsEllipseItem(-4, -4, 6, 6, self)
        self.innerCircle.setPen(QPen(QColor(0, 0, 0, 0), 0))

        self.visibleColor = QColor(0, 0, 0)

        self.innerCircle.setBrush(self.visibleColor)

        self.setCursor(QCursor(QtCore.Qt.CrossCursor))
        # self.setCursor(QCursor(QtCore.Qt.PointingHandCursor))

        # The Port itself is larger than only the innerCircle
        self.setBrush(QBrush(QtCore.Qt.white))

        # Hacky fix for no border of ellipse
        p1 = QPen(QColor(0, 0, 0, 0), 0)  # pylint: disable = invalid-name
        self.setPen(p1)

        self.setFlag(self.ItemSendsScenePositionChanges, True)
        self.setAcceptHoverEvents(True)

        # QUndo framework related
        self.savePos = None
        self.savedPos = False

    def setParent(self, p):  # pylint: disable = invalid-name
        self.parent = p
        self.setParentItem(p)

    def itemChange(self, change, value):
        # pylint: disable = fixme, too-many-branches, too-many-statements, too-many-nested-blocks
        # TODO : here to merge segments when moving blockitems
        if (
                self.parent.parent.parent().moveDirectPorts
                and hasattr(self.parent, "heatExchangers")
                and change == self.ItemPositionChange
        ):
            if not self.savePos is None:
                self.logger.debug("val is " + str(value))
                value.setY(max(value.y(), 0))
                value.setY(min(value.y(), self.parent.h))

                return QPointF(self.savePos.x(), value.y())

        if change == self.ItemScenePositionHasChanged and self.parent.parent.parent().editorMode == 0:
            self.logger.debug("editor mode = 0")
            for conn in self.connectionList:
                conn.positionLabel()

                if conn.fromPort is self:
                    e = conn.segments[0]  # pylint: disable = invalid-name
                    e.setLine(self.scenePos().x(), self.scenePos().y(), e.line().p2().x(), e.line().p2().y())

                if conn.toPort is self:
                    e = conn.segments[-1]  # pylint: disable = invalid-name
                    e.setLine(e.line().p1().x(), e.line().p1().y(), self.scenePos().x(), self.scenePos().y())

                for s in conn.segments:  # pylint: disable = invalid-name
                    s.updateGrad()

        if change == self.ItemScenePositionHasChanged and self.parent.parent.parent().editorMode == 1:
            for conn in self.connectionList:
                # Update position of connection label
                conn.positionLabel()

                if conn.fromPort is self:
                    if (self.createdAtSide not in (1, 3)) or not conn.segments[0].isVertical():
                        if len(conn.getCorners()) > 0 and len(conn.segments) > 0:
                            self.logger.debug("inside here")
                            cor = conn.getCorners()[0]
                            cor.setPos(cor.pos().x(), self.scenePos().y())

                            seg = conn.segments[0]  # first segment
                            seg.setLine(
                                self.scenePos().x(), self.scenePos().y(), cor.scenePos().x() + 0.6, cor.scenePos().y()
                            )
                            if len(conn.segments) > 2:
                                verSeg = conn.segments[1]
                                nextSeg = conn.segments[2]
                                if nextSeg.isHorizontal() and seg.isHorizontal():
                                    if (
                                            int(seg.endNode.parent.pos().y() - 0)
                                            <= int(nextSeg.line().p2().y())
                                            <= int(seg.endNode.parent.pos().y() + 0)
                                    ):
                                        self.logger.debug("both segments are horizontal from fromport")
                                        self.hideCorners(conn)
                                        verSeg.setVisible(False)
                                    else:
                                        self.showCorners(conn)
                                        verSeg.setVisible(True)
                    else:
                        self.logger.debug("inside else")
                        if len(conn.getCorners()) > 0 and len(conn.segments) > 0:
                            cor = conn.getCorners()[0]
                            cor.setPos(self.scenePos().x(), cor.pos().y())

                            seg = conn.segments[0]  # first segment
                            seg.setLine(self.scenePos().x(), self.scenePos().y(), cor.scenePos().x(),
                                        cor.scenePos().y())

                elif conn.toPort is self:
                    if (conn.fromPort.createdAtSide not in (1, 3)) or not conn.segments[0].isVertical():
                        if len(conn.getCorners()) > 0 and len(conn.segments) > 0:
                            cor = conn.getCorners()[-1]
                            cor.setPos(cor.pos().x(), self.scenePos().y())

                            seg = conn.segments[-1]
                            seg.setLine(
                                self.scenePos().x(),
                                self.scenePos().y(),
                                cor.scenePos().x() + 0.6,
                                cor.scenePos().y() + 0.6,
                            )
                            if len(conn.segments) > 2:
                                verSeg = conn.segments[-2]
                                nextSeg = conn.segments[-3]
                                if nextSeg.isHorizontal() and seg.isHorizontal():
                                    if (
                                            int(nextSeg.endNode.parent.pos().y() - 0)
                                            <= int(seg.line().p2().y())
                                            <= int(nextSeg.endNode.parent.pos().y() + 0)
                                    ):
                                        self.logger.debug("both segments are horizontal from toport")
                                        self.hideCorners(conn)
                                        verSeg.setVisible(False)
                                    else:
                                        self.showCorners(conn)
                                        verSeg.setVisible(True)
                    else:
                        if len(conn.getCorners()) == 1 and len(conn.segments) > 0:
                            cor = conn.getCorners()[-1]
                            cor.setPos(cor.pos().x(), self.scenePos().y())
                            self.logger.debug("Inside 2nd")

                            seg = conn.segments[-1]  # last segment
                            seg.setLine(self.scenePos().x(), self.scenePos().y(), cor.scenePos().x(),
                                        cor.scenePos().y())
                        elif len(conn.getCorners()) == 2 and len(conn.segments) > 0:
                            cor = conn.getCorners()[-1]
                            cor.setPos(self.scenePos().x(), cor.pos().y())
                            self.logger.debug("Inside 3rd")

                            seg = conn.segments[-1]  # last segment
                            seg.setLine(self.scenePos().x(), self.scenePos().y(), cor.scenePos().x(),
                                        cor.scenePos().y())

                else:
                    self.logger.debug("Error: In Mode 1, moving a portItem, portItem is neither from nor toPort")

        if change == self.ItemScenePositionHasChanged:
            for cb in self.posCallbacks:  # pylint: disable = invalid-name
                cb(value)
            return value
        return super().itemChange(change, value)

    def mousePressEvent(self, event):  # pylint: disable = unused-argument
        if self.parent.parent.parent().moveDirectPorts and hasattr(self.parent, "heatExchangers"):
            self.setFlag(self.ItemIsMovable)
            self.savePos = self.pos()
            return

        if self.connectionList:
            return

        self.setFlag(self.ItemIsMovable, False)
        self.scene().parent().startConnection(self)

    def hoverEnterEvent(self, event):  # pylint: disable = unused-argument
        self.logger.debug(self.parent)

        self.enlargePortSize()

        self.innerCircle.setBrush(self.ashColorB)

        self._highlightInternallyConnectedPortItems()

        self.debugprint()

    def enlargePortSize(self):
        if not self.savedPos:
            self.setPos(self.pos().x() - 3, self.pos().y() - 3)
            self.savedPos = True
        self.setRect(-4, -4, 10, 10)
        self.innerCircle.setRect(-4, -4, 10, 10)

    def hoverLeaveEvent(self, event):  # pylint: disable = unused-argument
        self.resetPortSize()

        self.innerCircle.setBrush(self.visibleColor)

        self._unhighlightInternallyConnectedPortItems()

        self._debugClear()

    def resetPortSize(self):
        if self.savedPos:
            self.setPos(self.pos().x() + 3, self.pos().y() + 3)
            self.savedPos = False
        self.setRect(-4, -4, 7, 7)
        self.innerCircle.setRect(-4, -4, 6.5, 6.5)

    def debugprint(self):

        self.parent.parent.parent().listV.addItem("ID: " + str(self.id))

        internalPiping = self.parent.getInternalPiping()
        portItemsAndInternalRealNode = internalPiping.getPortItemsAndAdjacentRealNodeForGraphicalPortItem(self)
        portItems = [pr.portItem for pr in portItemsAndInternalRealNode]
        formattedPortItems = [f"{p.name} ({p.type.value})" for p in portItems]
        jointFormattedPortItems = "\n".join(formattedPortItems)
        self.parent.parent.parent().listV.addItem(f"Names: {jointFormattedPortItems}")

        self.parent.parent.parent().listV.addItem("Block: " + self.parent.displayName)

        self.parent.parent.parent().listV.addItem("Connections:")
        for connection in self.connectionList:
            self.parent.parent.parent().listV.addItem(connection.displayName)

    def _debugClear(self):
        self.parent.parent.parent().listV.clear()

    def hideCorners(self, connection):  # pylint: disable = no-self-use
        cor = connection.getCorners()[0]
        cor2 = connection.getCorners()[-1]
        cor.setVisible(False)
        cor2.setVisible(False)

    def showCorners(self, connection):  # pylint: disable = no-self-use
        cor = connection.getCorners()[0]
        cor2 = connection.getCorners()[-1]
        cor.setVisible(True)
        cor2.setVisible(True)

    def getConnectedRealNode(
            self,
            portItem: _mfn.PortItem,
            massFlowContributor: _mfs.MassFlowNetworkContributorMixin,
    ) -> _tp.Optional[_mfn.RealNodeBase]:
        connectedMassFlowContributor = self._getConnectedMassFlowContributor(massFlowContributor)
        if not connectedMassFlowContributor:
            return None

        connectedInternalPiping = connectedMassFlowContributor.getInternalPiping()

        connectedPortItemsAndAdjacentRealNode = (
            connectedInternalPiping.getPortItemsAndAdjacentRealNodeForGraphicalPortItem(self)
        )

        selectedConnectedRealNode = self._selectConnectedRealNode(portItem, connectedPortItemsAndAdjacentRealNode)

        return selectedConnectedRealNode

    def _getConnectedMassFlowContributor(
            self, massFlowContributor: _mfs.MassFlowNetworkContributorMixin
    ) -> _tp.Optional[_mfs.MassFlowNetworkContributorMixin]:
        if massFlowContributor == self.parent:
            return self.getConnectionOrNone()

        connection = self.getConnectionOrNone()

        if not connection or connection != massFlowContributor:
            raise ValueError("`massFlowContributor' is not adjacent to this port item.")

        return self.parent

    def getConnectionOrNone(self) -> _tp.Optional[_cb.ConnectionBase]:
        if not self.connectionList:
            return None

        assert len(self.connectionList) == 1

        connection = self.connectionList[0]

        return connection

    def getConnection(self) -> _cb.ConnectionBase:
        connection = self.getConnectionOrNone()

        assert connection

        return connection

    def _selectConnectedRealNode(
            self,
            portItem: _mfn.PortItem,
            connectedPortItemsAndAdjacentRealNode: _tp.Sequence[_mfs.PortItemAndAdjacentRealNode],
    ) -> _mfn.RealNodeBase:
        raise NotImplementedError()

    def _highlightInternallyConnectedPortItems(self):
        raise NotImplementedError()

    def _unhighlightInternallyConnectedPortItems(self):
        raise NotImplementedError()
